import logging

import numpy as np
from astropy import coordinates
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.modeling import models
from gwcs import WCS, FITSImagingWCSTransform, coordinate_frames
from stcal.alignment.util import (
    calc_rotation_matrix,
    compute_s_region_keyword,
    compute_scale,
)

from romancal.assign_wcs.utils import create_footprint

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def assign_l3_wcs(model, wcs):
    """Assign a wcs to a level 3 model and update the wcsinfo

    Parameters
    ----------
    model : `DataModel`
        The model whose meta is to be updated.

    wcs : `GWCS`
        GWCS info to transfer into the `meta.wcsinfo` block

    Notes
    -----
    Some models/parameters in the GWCS object have explicit names, such as
    'crpix1'. However, some do not and hence have to be accessed explicitly
    by indexing. This is fragile and will be a source of issues.
    """
    model.meta.wcs = wcs
    l3_wcsinfo = model.meta.wcsinfo
    transform = wcs.forward_transform

    l3_wcsinfo.projection = "TAN"
    l3_wcsinfo.image_shape = model.shape

    # Fill out image-local information
    pixel_center = [(v - 1) / 2.0 for v in model.shape[::-1]]
    world_center = wcs(*pixel_center)
    l3_wcsinfo.ra = world_center[0]
    l3_wcsinfo.dec = world_center[1]

    l3_wcsinfo.orientation = calc_pa(wcs, *world_center)

    footprint = create_footprint(wcs, model.shape, center=False)
    l3_wcsinfo.s_region = compute_s_region_keyword(footprint)

    try:
        l3_wcsinfo.x_ref = transform.crpix[0]
        l3_wcsinfo.y_ref = transform.crpix[1]
        world_ref = transform.crval
    except AttributeError:
        log.warning(
            "WCS has no clear reference pixel defined by crpix1/crpix2. Assuming reference pixel is center."
        )
        l3_wcsinfo.x_ref = pixel_center[0]
        l3_wcsinfo.y_ref = pixel_center[1]

        world_ref = wcs(l3_wcsinfo.x_ref, l3_wcsinfo.y_ref, with_bounding_box=False)
    l3_wcsinfo.ra_ref = world_ref[0]
    l3_wcsinfo.dec_ref = world_ref[1]

    try:
        l3_wcsinfo.pixel_scale_ref = transform.cdelt[0]
    except AttributeError:
        l3_wcsinfo.pixel_scale_ref = compute_scale(wcs, world_ref)

    # pixel scale at center of image
    l3_wcsinfo.pixel_scale = compute_scale(wcs, world_center)

    l3_wcsinfo.orientation_ref = calc_pa(wcs, *world_ref)

    try:
        l3_wcsinfo.rotation_matrix = transform.pc.value.tolist()
    except AttributeError:
        log.warning(
            "WCS has no clear rotation matrix defined by pc_rotation_matrix. Calculating one."
        )
        rotation_matrix = calc_rotation_matrix(
            np.deg2rad(l3_wcsinfo.orientation_ref), 0.0
        )
        l3_wcsinfo.rotation_matrix = rotation_matrix


def calc_pa(wcs, ra, dec):
    """Calculate position angle at given ra,dec

    Parameters
    ----------
    wcs : GWCS
        The wcs in consideration.

    ra, dec : float, float
        The ra/dec in degrees.

    Returns
    -------
    position_angle : float
        The position angle in degrees.

    """
    delta_pix = [v for v in wcs.invert(ra, dec, with_bounding_box=False)]
    delta_pix[1] += 100
    delta_coord = SkyCoord(
        *wcs(*delta_pix, with_bounding_box=False), frame="icrs", unit="deg"
    )
    coord = SkyCoord(ra, dec, frame="icrs", unit="deg")

    return coord.position_angle(delta_coord).degree


def l3wcsinfo_to_wcs(wcsinfo, bounding_box=None):
    crpix = [wcsinfo["x_ref"], wcsinfo["y_ref"]]
    cdelt = [wcsinfo["pixel_scale_ref"], wcsinfo["pixel_scale_ref"]]
    tangent_projection = models.Pix2Sky_TAN()
    crval = [
        wcsinfo["ra_ref"],
        wcsinfo["dec_ref"],
    ]
    pc = np.array(wcsinfo["rotation_matrix"])
    det2sky = FITSImagingWCSTransform(
        tangent_projection, crpix=crpix, crval=crval, cdelt=cdelt, pc=pc
    )
    det2sky.name = "detector_to_sky"

    detector_frame = coordinate_frames.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = coordinate_frames.CelestialFrame(
        reference_frame=coordinates.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )
    wcsobj = WCS(
        [(detector_frame, det2sky), (sky_frame, None)], name=wcsinfo.get("name", None)
    )

    if bounding_box:
        wcsobj.bounding_box = bounding_box

    return wcsobj



"""Create an association based on skycells"""

import argparse
import logging
import os
import sys

import numpy as np
import roman_datamodels as rdm

import romancal.skycell.match as sm
import romancal.skycell.skymap as sc
from romancal.associations import asn_from_list
from romancal.associations.lib.utilities import mk_level3_asn_name
from romancal.lib.basic_utils import parse_visitID as parse_visitID

__all__ = ["skycell_asn"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.setLevel("INFO")


def skycell_asn(
    filelist: list[str],
    output_file_root: os.PathLike,
    product_type: str,
    release_product: str,
):
    """
    Create the skycell association from a list of L2 calibrated files.

    This function processes a list of Level 2 calibrated files, identifies matching skycells, generates
    TAN WCS parameters, and creates an association file for the identified skycells.

    Parameters
    ----------
    filelist : list of str
        List of file names to be processed.
    output_file_root : str
        Root string for the output association file.
    product_type : str
        Type of product when creating the association (e.g., 'visit', 'daily').
    release_product : str
        Release product identifier for the association.

    Returns
    -------
    None
    """
    skycell_indices = []
    file_list = []
    for file_name in filelist:
        cal_file = rdm.open(file_name)
        filter_id = cal_file.meta.instrument.optical_element.lower()
        intersecting_skycells = sm.find_skycell_matches(cal_file.meta.wcs)
        logger.info(f"Skycell List:{file_name}, {intersecting_skycells}")
        file_list.append([file_name, intersecting_skycells])
        skycell_indices.extend(intersecting_skycells)

    unique_skycell_indices = np.unique(skycell_indices)
    for skycell_index in unique_skycell_indices:
        member_list = []
        skycell = sc.SkyCell(skycell_index)
        for a in file_list:
            if np.isin(skycell_index, a[1]):
                member_list.append(a[0])

        # grab all the wcs parameters needed for generate_tan_wcs
        parsed_visit_id = parse_visitID(member_list[0][1:20])
        program_id = parsed_visit_id["Program"]
        asn_file_name = mk_level3_asn_name(
            member_list[0][1:20],
            output_file_root,
            filter_id,
            release_product,
            product_type,
            skycell.name,
        )

        prompt_product_asn = asn_from_list.asn_from_list(
            member_list, product_name=asn_file_name
        )
        prompt_product_asn["asn_type"] = "image"
        prompt_product_asn["program"] = program_id
        prompt_product_asn["target"] = skycell.name
        prompt_product_asn["skycell_wcs_info"] = skycell.wcs_info

        _, serialized = prompt_product_asn.dump(format="json")

        with open(asn_file_name + "_asn.json", "w") as outfile:
            outfile.write(serialized)


def _cli(args=None):
    """Command-line interface for list_to_asn

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """
    if args is None:
        args = sys.argv[1:]
    if isinstance(args, str):
        args = args.split(" ")

    parser = argparse.ArgumentParser(
        description="Create an association from a list of files",
        usage="skycell_asn --product-type visit --release-product prompt *_cal.asdf -o r512",
    )

    parser.add_argument(
        "-o",
        "--output-file-root",
        type=str,
        required=True,
        help="Root string for file to write association to",
    )

    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="json",
        help='Format of the association files. Default: "%(default)s"',
    )

    parser.add_argument(
        "--product-type",
        type=str,
        default="visit",
        help="The product type when creating the association",
    )

    parser.add_argument(
        "--release-product",
        type=str,
        default="p",
        help="The release product when creating the association",
    )

    parser.add_argument(
        "-r",
        "--rule",
        type=str,
        default="DMS_ELPP_Base",
        help=('The rule to base the association structure on. Default: "%(default)s"'),
    )
    parser.add_argument(
        "-i",
        "--id",
        type=str,
        default="o999",
        help='The association candidate id to use. Default: "%(default)s"',
        dest="acid",
    )

    parser.add_argument(
        "filelist",
        type=str,
        nargs="+",
        help="File list to include in the association",
    )

    parsed = parser.parse_args(args=args)
    logger.info("Command-line arguments: %s", parsed)

    skycell_asn(
        parsed.filelist,
        parsed.output_file_root,
        parsed.product_type,
        parsed.release_product,
    )



import asdf
import numpy as np
import pytest
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling import models
from astropy.time import Time
from gwcs import WCS
from gwcs import coordinate_frames as cf
from roman_datamodels import datamodels
from stcal.resample.utils import compute_mean_pixel_area

from romancal.assign_wcs.utils import add_s_region
from romancal.datamodels import ModelLibrary
from romancal.lib.tests.helpers import word_precision_check
from romancal.resample import ResampleStep, resample_utils


class WfiSca:
    def __init__(
        self,
        fiducial_world,
        pscale=(0.000031, 0.000031),
        shape=(100, 100),
        filename="noname",
    ):
        self.fiducial_world = fiducial_world
        self.pscale = pscale
        self.shape = shape
        self.filename = filename

    def create_image(self):
        """
        Create a dummy L2 datamodel given the coordinates of the fiducial point,
        a pixel scale, and the image shape and filename.

        Returns
        -------
        datamodels.ImageModel
            An L2 ImageModel datamodel.
        """
        rng = np.random.default_rng(seed=13)
        l2 = datamodels.ImageModel.create_fake_data(
            {
                "meta": {
                    "wcsinfo": {
                        "ra_ref": 10,
                        "dec_ref": 0,
                        "vparity": -1,
                        "v3yangle": -60,
                    },
                    "exposure": {
                        "exposure_time": 152.04000000000002,
                        "effective_exposure_time": 3.04 * 6 * 8,
                    },
                    "observation": {
                        "program": 5,
                        "execution_plan": 1,
                        "pass": 1,
                        "observation": 1,
                        "segment": 1,
                        "visit": 1,
                        "visit_file_group": 1,
                        "visit_file_sequence": 1,
                        "visit_file_activity": "01",
                        "exposure": 1,
                    },
                    "photometry": {
                        "pixel_area": None,
                    },
                },
                "data": rng.poisson(2.5, size=self.shape).astype(np.float32),
                "var_rnoise": rng.normal(1, 0.05, size=self.shape).astype(np.float32),
                "var_poisson": rng.poisson(1, size=self.shape).astype(np.float32),
                "var_flat": rng.uniform(0, 1, size=self.shape).astype(np.float32),
            },
            shape=self.shape,
        )
        # data from WFISim simulation of SCA #01
        l2.meta.filename = self.filename
        l2.meta["wcs"] = create_wcs_object_without_distortion(
            fiducial_world=self.fiducial_world,
            pscale=self.pscale,
            shape=self.shape,
        )
        l2.meta.photometry.pixel_area = compute_mean_pixel_area(l2.meta.wcs, self.shape)
        model = datamodels.ImageModel(l2)
        add_s_region(model)
        return model


def create_wcs_object_without_distortion(fiducial_world, pscale, shape):
    """
    Create a simple WCS object without either distortion or rotation.

    Parameters
    ----------
    fiducial_world : tuple
        A pair of values corresponding to the fiducial's world coordinate.
    pscale : tuple
        A pair of values corresponding to the pixel scale in each axis.
    shape : tuple
        A pair of values specifying the dimensions of the WCS object.

    Returns
    -------
    gwcs.WCS
        A gwcs.WCS object.
    """
    # components of the model
    shift = models.Shift() & models.Shift()

    affine = models.AffineTransformation2D(
        matrix=[[1, 0], [0, 1]], translation=[0, 0], name="pc_rotation_matrix"
    )

    scale = models.Scale(pscale[0]) & models.Scale(pscale[1])

    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        fiducial_world[0],
        fiducial_world[1],
        180,
    )

    det2sky = shift | affine | scale | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.FK5(), name="fk5", unit=(u.deg, u.deg)
    )

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]

    wcs_obj = WCS(pipeline)

    wcs_obj.bounding_box = (
        (-0.5, shape[-1] - 0.5),
        (-0.5, shape[-2] - 0.5),
    )

    wcs_obj.pixel_shape = shape[::-1]
    wcs_obj.array_shape = shape

    return wcs_obj


def _wfi_sca1():
    sca = WfiSca(
        fiducial_world=(10, 0),
        filename="r0000501001001001001_01101_0001_wfi01_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca1():
    return _wfi_sca1()


def _wfi_sca2():
    sca = WfiSca(
        fiducial_world=(10.00139, 0),
        filename="r0000501001001001001_01101_0001_wfi02_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca2():
    return _wfi_sca2()


def _wfi_sca3():
    sca = WfiSca(
        fiducial_world=(10.00278, 0),
        filename="r0000501001001001001_01101_0001_wfi03_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca3():
    return _wfi_sca3()


def _wfi_sca4():
    sca = WfiSca(
        fiducial_world=(10, 0),
        filename="r0000501001001001001_01101_0002_wfi01_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca4():
    return _wfi_sca4()


def _wfi_sca5():
    sca = WfiSca(
        fiducial_world=(10.00139, 0),
        filename="r0000501001001001001_01101_0002_wfi02_cal.asdf",
    )
    return sca.create_image()


@pytest.fixture
def wfi_sca5():
    return _wfi_sca5()


def _wfi_sca6():
    sca = WfiSca(
        fiducial_world=(10.00278, 0),
        filename="r0000501001001001001_01101_0002_wfi03_cal.asdf",
    )

    return sca.create_image()


@pytest.fixture
def wfi_sca6():
    return _wfi_sca6()


def _exposure_1(wfi_sca1, wfi_sca2, wfi_sca3):
    """Returns a list with models corresponding to a dummy exposure 1."""
    # set the same exposure time for all SCAs
    for sca in [wfi_sca1, wfi_sca2, wfi_sca3]:
        sca.meta.exposure["start_time"] = Time(
            "2020-02-01T00:00:00", format="isot", scale="utc"
        )
        sca.meta.exposure["end_time"] = Time(
            "2020-02-01T00:02:30", format="isot", scale="utc"
        )
        sca.meta.observation.exposure = 1
        sca.meta.observation.observation_id = "1"
    return [wfi_sca1, wfi_sca2, wfi_sca3]


@pytest.fixture
def exposure_1(wfi_sca1, wfi_sca2, wfi_sca3):
    return _exposure_1(wfi_sca1, wfi_sca2, wfi_sca3)


def _exposure_2(wfi_sca4, wfi_sca5, wfi_sca6):
    """Returns a list with models corresponding to a dummy exposure 2."""
    # set the same exposure time for all SCAs
    for sca in [wfi_sca4, wfi_sca5, wfi_sca6]:
        sca.meta.exposure["start_time"] = Time(
            "2020-05-01T00:00:00", format="isot", scale="utc"
        )
        sca.meta.exposure["end_time"] = Time(
            "2020-05-01T00:02:30", format="isot", scale="utc"
        )
        sca.meta.observation.exposure = 2
        sca.meta.observation.observation_id = "2"
    return [wfi_sca4, wfi_sca5, wfi_sca6]


@pytest.fixture
def exposure_2(wfi_sca4, wfi_sca5, wfi_sca6):
    """Returns a list with models corresponding to a dummy exposure 2."""
    return _exposure_2(wfi_sca4, wfi_sca5, wfi_sca6)


@pytest.fixture
def multiple_exposures(exposure_1, exposure_2):
    """Returns a list with all the datamodels from exposure 1 and 2."""
    exposure_1.extend(exposure_2)
    return exposure_1


def get_resampled_wcs_pixel_scale(wcs):
    return wcs.forward_transform.cdelt


def test_resampledata_do_drizzle_many_to_one_default_no_rotation_single_exposure(
    exposure_1,
):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that its extrema fall within the output WCS footprint.

    N.B.: since we are not providing the rotation parameter for the
    resample_utils.make_output_wcs method, the output WCS will have
    the same orientation (i.e. same PA) as the detector axes.
    """

    input_models = ModelLibrary(exposure_1)
    output_model = ResampleStep().run(input_models)

    output_min_value = np.min(output_model.meta.wcs.footprint())
    output_max_value = np.max(output_model.meta.wcs.footprint())

    def get_footprint(model, index):
        return model.meta.wcs.footprint()

    input_wcs_list = list(input_models.map_function(get_footprint, modify=False))

    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    np.testing.assert_array_less(output_min_value, expected_min_value)
    np.testing.assert_(output_max_value > expected_max_value)


def test_resampledata_do_drizzle_many_to_one_default_no_rotation_multiple_exposures(
    multiple_exposures,
):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that its extrema fall within the output WCS footprint.

    N.B.: since we are not providing the rotation parameter for the
    resample_utils.make_output_wcs method, the output WCS will have
    the same orientation (i.e. same PA) as the detector axes.
    """

    input_models = ModelLibrary(multiple_exposures)
    output_model = ResampleStep().run(input_models)

    output_min_value = np.min(output_model.meta.wcs.footprint())
    output_max_value = np.max(output_model.meta.wcs.footprint())

    def get_footprint(model, index):
        return model.meta.wcs.footprint()

    input_wcs_list = list(input_models.map_function(get_footprint, modify=False))

    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    np.testing.assert_array_less(output_min_value, expected_min_value)
    np.testing.assert_(output_max_value > expected_max_value)


def test_resampledata_do_drizzle_many_to_one_default_rotation_0(exposure_1):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that the output WCS footprint vertices are close to the
    expected vertices for the combined input WCS footprint.

    N.B.: in this case, rotation=0 will create a WCS that will be oriented North up.
    """

    input_models = ModelLibrary(exposure_1)

    output_model = ResampleStep(rotation=0).run(input_models)

    pscale = get_resampled_wcs_pixel_scale(output_model.meta.wcs)
    output_min_value = np.min(output_model.meta.wcs.footprint())
    output_max_value = np.max(output_model.meta.wcs.footprint())

    def get_footprint(model, index):
        return model.meta.wcs.footprint()

    input_wcs_list = list(input_models.map_function(get_footprint, modify=False))

    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    assert (
        (expected_min_value - 0.5001 * pscale[0])
        <= output_min_value
        <= expected_min_value
    )
    assert (
        (expected_max_value + 0.5001 * pscale[0])
        >= output_max_value
        >= expected_max_value
    )


def test_resampledata_do_drizzle_many_to_one_default_rotation_0_multiple_exposures(
    multiple_exposures,
):
    """Test that output WCS encompass the entire combined input WCS region
    by checking that the output WCS footprint vertices are close to the
    expected vertices for the combined input WCS footprint.

    N.B.: in this case, rotation=0 will create a WCS that will be oriented North up.
    """

    input_models = ModelLibrary(multiple_exposures)
    output_model = ResampleStep(rotation=0).run(input_models)

    output_min_value = np.min(output_model.meta.wcs.footprint())
    output_max_value = np.max(output_model.meta.wcs.footprint())
    pscale = get_resampled_wcs_pixel_scale(output_model.meta.wcs)

    def get_footprint(model, index):
        return model.meta.wcs.footprint()

    input_wcs_list = list(input_models.map_function(get_footprint, modify=False))

    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    # Assert
    assert (
        (expected_min_value - 0.5001 * pscale[0])
        <= output_min_value
        <= expected_min_value
    )
    assert (
        (expected_max_value + 0.5001 * pscale[0])
        >= output_max_value
        >= expected_max_value
    )


def test_resampledata_do_drizzle_many_to_one_single_input_model(wfi_sca1):
    """Test that the output of resample from a single input file creates a WCS
    footprint vertices that are close to the input WCS footprint's vertices."""

    input_models = ModelLibrary([wfi_sca1])
    output_model = ResampleStep(rotation=0).run(input_models)

    flat_1 = np.sort(wfi_sca1.meta.wcs.footprint().flatten())
    flat_2 = np.sort(output_model.meta.wcs.footprint().flatten())
    pscale = get_resampled_wcs_pixel_scale(output_model.meta.wcs)

    np.testing.assert_allclose(flat_1, flat_2, atol=0.5 * pscale[0])


def test_update_exposure_times_different_sca_same_exposure(exposure_1):
    """Test that update_exposure_times is properly updating the exposure parameters
    for a set of different SCAs belonging to the same exposure."""
    input_models = ModelLibrary(exposure_1)
    output_model = ResampleStep().run(input_models)

    # these three SCAs overlap, so the max exposure time is 3x.
    # get this time within 0.1 s.
    time_difference = (
        output_model.meta.coadd_info.max_exposure_time
        - 3 * exposure_1[0].meta.exposure.effective_exposure_time
    )
    assert np.abs(time_difference) < 0.1
    assert (
        output_model.meta.coadd_info.time_first
        == exposure_1[0].meta.exposure.start_time
    )
    assert (
        output_model.meta.coadd_info.time_last == exposure_1[0].meta.exposure.end_time
    )


def test_update_exposure_times_same_sca_different_exposures(exposure_1, exposure_2):
    """Test that update_exposure_times is properly updating the exposure parameters
    for a set of the same SCA but belonging to different exposures."""
    input_models = ModelLibrary([exposure_1[0], exposure_2[0]])
    output_model = ResampleStep().run(input_models)

    with input_models:
        models = list(input_models)
        first_time = min(x.meta.exposure.start_time for x in models)
        last_time = max(x.meta.exposure.end_time for x in models)
        [input_models.shelve(model, i, modify=False) for i, model in enumerate(models)]

    # these exposures overlap perfectly so the max exposure time should
    # be equal to the individual time times two.
    time_difference = (
        output_model.meta.coadd_info.max_exposure_time
        - 2 * exposure_1[0].meta.exposure.effective_exposure_time
    )
    assert np.abs(time_difference) < 0.1

    assert output_model.meta.coadd_info.time_first == first_time

    assert output_model.meta.coadd_info.time_last == last_time

    # likewise the per-pixel median exposure time is just 2x the individual
    # sca exposure time.
    time_difference = (
        output_model.meta.coadd_info.max_exposure_time
        - 2 * exposure_1[0].meta.exposure.effective_exposure_time
    )
    assert np.abs(time_difference) < 0.1


@pytest.mark.parametrize("include_var_flat", [False])  # , True])
def test_var_flat_presence(exposure_1, include_var_flat):
    """Test that var_flat is included or excluded depending on its presence in the underlying exposures."""
    if not include_var_flat:
        exposure_1 = [e.copy() for e in exposure_1]
        for e in exposure_1:
            del e._instance["var_flat"]
            print(getattr(e, "var_flat", None))
    input_models = ModelLibrary(exposure_1)
    print("2")
    with input_models:
        models = list(input_models)
        for e in models:
            print(getattr(e, "var_flat", None))
        [input_models.shelve(e, modify=False) for e in models]
    print("1")
    output_model = ResampleStep().run(input_models)

    if not include_var_flat:
        assert not hasattr(output_model, "var_flat")
    else:
        assert hasattr(output_model, "var_flat")


def test_custom_wcs_input_small_overlap_no_rotation(wfi_sca1, wfi_sca3, tmp_path):
    """Test that resample can create a proper output in the edge case where the
    desired output WCS does not encompass the entire input datamodel but, instead, have
    just a small overlap."""
    input_models = ModelLibrary([wfi_sca1])
    wcs_path = tmp_path / "wcs.asdf"
    wcs3 = wfi_sca3.meta.wcs
    asdf.AsdfFile({"wcs": wcs3}).write_to(wcs_path)

    output_model = ResampleStep(output_wcs=str(wcs_path), rotation=0).run(input_models)

    np.testing.assert_allclose(output_model.meta.wcs(0, 0), wfi_sca3.meta.wcs(0, 0))


def test_custom_wcs_input_entire_field_no_rotation(multiple_exposures, tmp_path):
    """Test that resample can create a proper output that encompasses the entire
    combined FOV of the input datamodels."""
    input_models = ModelLibrary(multiple_exposures)

    output_wcs, _, _ = resample_utils.make_output_wcs(
        input_models,
        rotation=0,
    )

    wcs_path = tmp_path / "wcs.asdf"
    asdf.AsdfFile({"wcs": output_wcs}).write_to(wcs_path)
    output_model = ResampleStep(output_wcs=str(wcs_path)).run(input_models)

    output_min_value = np.min(output_model.meta.wcs.footprint())
    output_max_value = np.max(output_model.meta.wcs.footprint())
    pscale = get_resampled_wcs_pixel_scale(output_model.meta.wcs)

    def get_footprint(model, index):
        return model.meta.wcs.footprint()

    input_wcs_list = list(input_models.map_function(get_footprint, modify=False))

    expected_min_value = np.min(np.stack(input_wcs_list))
    expected_max_value = np.max(np.stack(input_wcs_list))

    assert (
        (expected_min_value - 0.5001 * pscale[0])
        <= output_min_value
        <= expected_min_value
    )
    assert (
        (expected_max_value + 0.5001 * pscale[0])
        >= output_max_value
        >= expected_max_value
    )


@pytest.mark.parametrize("weight_type", ["ivm", "exptime"])
def test_resampledata_do_drizzle_default_single_exposure_weight_array(
    exposure_1,
    weight_type,
):
    """Test that resample methods return non-empty weight arrays."""

    input_models = ModelLibrary(exposure_1)
    output_model = ResampleStep(weight_type=weight_type).run(input_models)
    assert np.any(output_model.weight > 0)


def test_l3_wcsinfo(multiple_exposures):
    """Test the population of the Level 3 wcsinfo block"""
    expected = {
        "ra_ref": 10.00292450000052,
        "dec_ref": 0.001534500000533253,
        "x_ref": 106.4579605214774,
        "y_ref": 80.66617532540977,
        "pixel_scale_ref": 3.100000000097307e-05,
        "pixel_scale": 3.099999999719185e-05,
        "image_shape": (161, 213),
        "ra": 10.002930353020417,
        "dec": 0.0015101325554100666,
        "orientation": 20.999999978134802,
        "orientation_ref": 20.99999999880985,
        "projection": "TAN",
        "s_region": (
            "POLYGON ICRS  10.005118262 -0.002002769 10.006906876 "
            "0.002656731 10.000742444 0.005023034 9.998953830 0.000363534"
        ),
    }

    input_models = ModelLibrary(multiple_exposures)
    output_model = ResampleStep().run(input_models)

    assert output_model.meta.wcsinfo.projection == expected["projection"]
    assert word_precision_check(
        output_model.meta.wcsinfo.s_region, expected["s_region"]
    )
    for key in expected.keys():
        if key not in ["projection", "s_region"]:
            assert np.allclose(output_model.meta.wcsinfo[key], expected[key])


def test_resample_pixel_scale_units(wfi_sca1):
    """
    Test that the input pixel_scale is in units of arcseconds.
    """
    pixel_scale = (0.05 * u.arcsec).value
    output_model = ResampleStep.call(wfi_sca1, pixel_scale=pixel_scale)
    coords = output_model.meta.wcs.pixel_to_world((100, 100), (100, 101))
    pscale = coords[0].separation(coords[1]).to(u.arcsec).value
    np.testing.assert_allclose(pscale, pixel_scale)



import numpy as np
import pytest
from roman_datamodels import datamodels as rdm

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_resample_single_file(
    rtdata, ignore_asdf_paths, resource_tracker, request, dms_logger
):
    output_data = "mosaic_resamplestep.asdf"

    rtdata.get_asn("WFI/image/L3_mosaic_asn.json")
    rtdata.get_truth(f"truth/WFI/image/{output_data}")

    rtdata.output = output_data

    args = [
        "romancal.step.ResampleStep",
        rtdata.input,
        "--rotation=0",
        "--resample_on_skycell=False",
        f"--output_file='{rtdata.output}'",
    ]
    with resource_tracker.track(log=request):
        RomanStep.from_cmdline(args)

    resample_out = rdm.open(rtdata.output)

    dms_logger.info(
        "ResampleStep recorded as complete? :"
        f" {resample_out.meta.cal_step.resample == 'COMPLETE'}"
    )
    assert resample_out.meta.cal_step.resample == "COMPLETE"

    dms_logger.info(
        "ResampleStep created 'meta.resample'? :"
        f" {hasattr(resample_out.meta, 'resample')}"
    )
    assert hasattr(resample_out.meta, "resample")

    dms_logger.info(
        f"""DMS342 MSG: Was ICRS used as the mosaic astrometric reference frame? :\
            {resample_out.meta.coordinates.reference_frame == "ICRS"}
        """
    )
    assert resample_out.meta.coordinates.reference_frame == "ICRS"

    dms_logger.info(
        f"""DMS343 MSG: ResampleStep created new attribute data quality information? :\
            {
            all(
                hasattr(resample_out, x)
                for x in [
                    "data",
                    "err",
                    "var_poisson",
                    "var_rnoise",
                ]
            )
        }"""
    )
    assert all(
        hasattr(resample_out, x) for x in ["data", "err", "var_poisson", "var_rnoise"]
    )

    dms_logger.info(
        f"""DMS343 MSG: Were the variance arrays populated (variance propagation)? :\
            {
            all(
                np.sum(~np.isnan(getattr(resample_out, x)))
                for x in [
                    "var_poisson",
                    "var_rnoise",
                ]
            )
        }"""
    )
    assert all(
        np.sum(~np.isnan(getattr(resample_out, x)))
        for x in ["var_poisson", "var_rnoise"]
    )

    dms_logger.info(
        f"""DMS343 MSG: Are there NaNs or zeros in the variance arrays, indicating poor data quality? :\
            {
            any(
                np.sum(
                    np.logical_or(
                        np.isnan(getattr(resample_out, x)),
                        np.equal(getattr(resample_out, x), 0),
                    )
                )
                > 0
                for x in ["var_poisson", "var_rnoise"]
            )
        }"""
    )
    assert all(
        np.sum(np.isnan(getattr(resample_out, x)))
        for x in ["var_poisson", "var_rnoise"]
    )

    dms_logger.info(
        f"""DMS344 MSG: ResampleStep created new attribute with total exposure time? :\
            {"max_exposure_time" in resample_out.meta.coadd_info}
        """
    )
    assert "max_exposure_time" in resample_out.meta.coadd_info

    dms_logger.info(
        f"""DMS345 MSG: ResampleStep included all metadata relevant to the creation of the mosaic? :\
            {
            all(
                hasattr(resample_out.meta.resample, x)
                and bool(getattr(resample_out.meta.resample, x))
                for x in [
                    "pixel_scale_ratio",
                    "pixfrac",
                    "pointings",
                    "weight_type",
                    "members",
                ]
            )
        }"""
    )
    assert all(
        hasattr(resample_out.meta.resample, x)
        and bool(getattr(resample_out.meta.resample, x))
        for x in [
            "pixel_scale_ratio",
            "pixfrac",
            "pointings",
            "weight_type",
            "members",
        ]
    )

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    dms_logger.info(f"Was the proper Resample data produced? : {diff.identical}")
    assert diff.identical, diff.report()



import math

import numpy as np
from stcal.alignment.util import compute_scale, wcs_from_sregions
from stcal.resample.utils import compute_mean_pixel_area


def make_output_wcs(
    input_models,
    pscale_ratio=1.0,
    pscale=None,
    rotation=None,
    shape=None,
    crpix: tuple[float, float] | None = None,
    crval: tuple[float, float] | None = None,
):
    """
    Generate output WCS here based on footprints of all input WCS objects

    Parameters
    ----------
    input_models : list of `roman_datamodels.datamodels.DataModel`
        Each datamodel must have a `gwcs.wcs.WCS` object.

    pscale_ratio : float, optional
        Ratio of output pixel scale to input pixel scale.
        Ignored when ``pscale`` is provided.

    pscale : float, None, optional
        Absolute pixel scale in degrees. When provided, overrides ``pscale_ratio``.

    rotation : float, None, optional
        Position angle (in degrees) of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes.

    shape : tuple of int, None, optional
        Shape of the image (data array) using ``numpy.ndarray`` convention
        (``ny`` first and ``nx`` second). This value will be assigned
        to ``pixel_shape`` and ``array_shape`` properties of the returned
        WCS object.

    crpix : tuple of float, None, optional
        Position of the reference pixel in the image array.  If ``ref_pixel`` is not
        specified, it will be set to the center of the bounding box of the
        returned WCS object.

    crval : tuple of float, None, optional
        Right ascension and declination of the reference pixel. Automatically
        computed if not provided.

    Returns
    -------
    output_wcs : object
        WCS object, with defined domain, covering entire set of input frames

    pscale : float
        The computed (or provided) input pixel scale (in degrees).

    pscale_ratio : float, optional
        The computed (or provided) input pixel scale ratio.
    """
    sregions = []
    ref_wcs = None
    ref_wcsinfo = None
    ref_shape = None
    with input_models:
        for model in input_models:
            if ref_wcs is None:
                ref_wcs = model.meta.wcs
                ref_wcsinfo = model.meta.wcsinfo
                ref_shape = model.data.shape
            sregions.append(model.meta.wcs.footprint())
            input_models.shelve(model, modify=False)

    if pscale is None:
        pscale = (
            compute_scale(
                ref_wcs,
                fiducial=np.array([ref_wcsinfo["ra_ref"], ref_wcsinfo["dec_ref"]]),
            )
            * pscale_ratio
        )
    else:
        pscale_ratio = pscale / np.rad2deg(
            math.sqrt(compute_mean_pixel_area(ref_wcs, shape=ref_shape))
        )

    wcs = wcs_from_sregions(
        sregions,
        ref_wcs=ref_wcs,
        ref_wcsinfo=ref_wcsinfo,
        pscale_ratio=pscale_ratio,
        pscale=pscale,
        shape=shape,
        rotation=rotation,
        crpix=crpix,
        crval=crval,
    )

    return wcs, pscale * 3600.0, pscale_ratio



#!/usr/bin/env python

"""V1 Calculation based on time and engineering database info."""

import argparse
import logging
from sys import stdout

from astropy.time import Time

import romancal.orientation.set_telescope_pointing as stp
from romancal.lib.engdb.engdb_tools import AVAILABLE_SERVICES
from romancal.orientation import v1_calculate

# Available reduce functions
REDUCE_FUNCS_MAPPING = {
    "all": stp.all_pointings,
    "first": stp.first_pointing,
    "average": stp.pointing_from_average,
}
REDUCE_FUNCS = list(REDUCE_FUNCS_MAPPING.keys())


# Begin execution
def main():
    """Calculate V1, the telescope's boresight axis, the direction the telescope is pointing."""
    parser = argparse.ArgumentParser(description="Calculate V1 over a time period.")

    parser.add_argument(
        "time_sources",
        type=str,
        nargs="+",
        help=(
            "Either a list of Roman data files to retrieve the timing from"
            " or a start and end time to retrieve pointing information for."
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        default=stdout,
        help="File to write V1 calculation table to. Default is standard output.",
    )
    parser.add_argument(
        "--pointing",
        type=str,
        choices=REDUCE_FUNCS,
        default="average",
        help=(
            "Which pointing(s) to use within the specified time interval."
            f" Choices: {REDUCE_FUNCS}"
            " (default: %(default)s)"
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity. Specifying multiple times adds more output.",
    )
    parser.add_argument(
        "--service",
        type=str,
        default="mast",
        choices=[name for name in AVAILABLE_SERVICES],
        help="Database service to use. Default: %(default)s",
    )

    # Arguments pertinent only to the EngdbMast service.
    if "mast" in AVAILABLE_SERVICES:
        parser.add_argument(
            "--engdb-url",
            type=str,
            default=None,
            help=(
                "URL of the engineering database."
                " If not specified, the environment variable 'ENG_BASE_URL' is used."
                " Otherwise, a hardwired default is used."
            ),
        )

    # Arguments pertinent only to the EngdbEDP service
    if "edp" in AVAILABLE_SERVICES:
        parser.add_argument(
            "--environment",
            type=str,
            default="test",
            choices=["dev", "test", "int", "ops"],
            help="Operational environment in use. Default: %(default)s",
        )
        parser.add_argument(
            "--path-to-cc",
            type=str,
            help="Full path to the required kerberos authentication keytab file",
        )

    args = parser.parse_args()

    # Set output detail.
    logger = logging.getLogger("romancal")
    logger_handler = logging.StreamHandler()
    logger.addHandler(logger_handler)
    logger_format_debug = logging.Formatter(
        "%(levelname)s:%(filename)s::%(funcName)s: %(message)s"
    )
    level = stp.LOGLEVELS[min(len(stp.LOGLEVELS) - 1, args.verbose)]
    logger.setLevel(level)
    if level <= logging.DEBUG:
        logger_handler.setFormatter(logger_format_debug)

    # Gather the service-specific args
    service_kwargs = {"service": args.service}
    for arg in ["engdb_url", "environment", "path_to_cc"]:
        try:
            service_kwargs[arg] = getattr(args, arg)
        except AttributeError:
            pass

    # Determine whether the sources are time specifications or a file list.
    if len(args.time_sources) == 2:
        try:
            obsstart = Time(args.time_sources[0])
            obsend = Time(args.time_sources[1])
        except ValueError:
            input_as_files = True
        else:
            logger.info(
                f"Retrieving V1 over the time span {obsstart.isot} - {obsend.isot}"
            )
            input_as_files = False
            if args.pointing != "all":
                logger.warning(
                    "V1 pointings have been requested over a time range. "
                    "However, the 'pointing' option is not 'all'."
                )
                logger.warning(
                    "There will only be a single result returned. Is this what was desired?"
                )
                logger.warning("Suggestion: Use '--pointing=all'")
    else:
        input_as_files = True

    # Process the file list.
    logger.info("Starting V1 calculation...")
    if input_as_files:
        v1s = v1_calculate.v1_calculate_from_models(
            args.time_sources,
            reduce_func=REDUCE_FUNCS_MAPPING[args.pointing],
            service_kwargs=service_kwargs,
        )
    else:
        v1s = v1_calculate.v1_calculate_over_time(
            obsstart,
            obsend,
            reduce_func=REDUCE_FUNCS_MAPPING[args.pointing],
            service_kwargs=service_kwargs,
        )

    formatted = v1_calculate.simplify_table(v1s)
    formatted.write(args.output, format="ascii.ecsv")
    logger.info("...V1 calculation completed.")


if __name__ == "__main__":
    main()



"""Roman tests for source catalog creation"""

import pytest
from astropy.table import Table

from romancal.stpipe import RomanStep

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(
    scope="module",
    params=[
        "r00001_p_v01001001001001_270p65x70y49_f158_coadd.asdf",
        "r0000101001001001001_f158_coadd.asdf",
        "r0000101001001001001_0001_wfi01_f158_cal.asdf",
    ],
    ids=["L3skycell", "L3", "L2"],
)
def run_source_catalog(rtdata_module, request, resource_tracker):
    rtdata = rtdata_module

    inputfn = request.param

    outputfn = inputfn.rsplit("_", 1)[0] + "_cat.parquet"
    rtdata.output = outputfn

    rtdata.get_data(f"WFI/image/{inputfn}")
    rtdata.input = inputfn
    rtdata.get_truth(f"truth/WFI/image/{outputfn}")

    args = [
        "romancal.step.SourceCatalogStep",
        rtdata.input,
    ]
    with resource_tracker.track():
        RomanStep.from_cmdline(args)
    return rtdata_module


@pytest.fixture(scope="module")
def catalog(run_source_catalog):
    yield Table.read(run_source_catalog.output)


@pytest.fixture(scope="module")
def fields(catalog):
    return catalog.dtype.names


@pytest.mark.parametrize(
    "field",
    (
        "ra_centroid",  # DMS374 positions on ICRF
        "dec_centroid",  # DMS374 positions on ICRF
        "segment_flux",  # DMS375 fluxes
        "kron_flux",  # DMS375 fluxes
        "aper02_flux",  # DMS399 aperture fluxes
        "aper04_flux",  # DMS399 aperture fluxes
        "aper08_flux",  # DMS399 aperture fluxes
        "aper02_flux_err",  # DMS386 flux uncertainties
        "segment_flux_err",  # DMS386 flux uncertainties
        "is_extended",  # DMS376 type of source
        "warning_flags",  # DMS387 dq_flags
    ),
)
def test_has_field(fields, field):
    assert field in fields


def test_log_tracked_resources(log_tracked_resources, run_source_catalog):
    log_tracked_resources()


def test_forced_catalog(rtdata_module, dms_logger):
    rtdata = rtdata_module
    input_deep_segm = "r00001_p_v01001001001001_270p65x70y49_f158_segm.asdf"
    input_shallow_coadd = "r00001_p_e01001001001001_0001_270p65x70y49_f158_coadd.asdf"
    truth_cat = "r00001_p_v01001001001001_270p65x70y49_f158_cat.parquet"
    rtdata.get_data(f"WFI/image/{input_deep_segm}")
    rtdata.get_data(f"WFI/image/{input_shallow_coadd}")
    truth_cat = rtdata.get_truth(f"truth/WFI/image/{truth_cat}")
    rtdata.input = input_shallow_coadd
    outputfn = input_shallow_coadd.rsplit("_", 1)[0] + "_force_cat.parquet"
    rtdata.output = outputfn

    args = [
        "romancal.step.SourceCatalogStep",
        rtdata.input,
        "--forced_segmentation",
        input_deep_segm,
        "--output_file",
        outputfn,
    ]
    RomanStep.from_cmdline(args)

    cat = Table.read(outputfn)
    fieldlist = [
        "forced_kron_flux",
        "forced_segment_flux",
        "forced_aper02_flux",
        "forced_semimajor",
        "forced_semiminor",
        "forced_ellipticity",
    ]
    for field in fieldlist:
        assert field in cat.dtype.names

    dms_logger.info(
        "DMS397: source catalog includes fields: "
        + ", ".join(fieldlist)
        + ", indicating measurements of morphology and photometry "
        "at multiple epochs."
    )

    cattruth = Table.read(truth_cat)
    assert set(cattruth.dtype.names) == set(cattruth.dtype.names)
    # weak assertion that our truth file must at least have the same
    # catalog fields as the file produced here.  Exactly matching rows
    # would require a lot of okifying things that aren't obviously
    # the same, but we can easily check that the columns match up
    # by name.



import copy
import getpass
import json
import os
from datetime import datetime

import pytest
from ci_watson.artifactory_helpers import UPLOAD_SCHEMA

from romancal.regtest.regtestdata import RegtestData
from romancal.regtest.resource_tracker import ResourceTracker

TODAYS_DATE = datetime.now().strftime("%Y-%m-%d")


@pytest.fixture(scope="session")
def artifactory_repos(pytestconfig):
    """Provides Artifactory inputs_root and results_root"""
    inputs_root = pytestconfig.getini("inputs_root")
    if not inputs_root:
        inputs_root = "roman-pipeline"
    else:
        inputs_root = inputs_root[0]

    results_root = pytestconfig.getini("results_root")
    if not results_root:
        results_root = "roman-pipeline-results/regression-tests/runs/"
    else:
        results_root = results_root[0]
    return inputs_root, results_root


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Create request.node.report_setup and request.node.report_call to be
    used in generate_artifactory_json() fixture.
    """
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "report_" + rep.when, rep)


def postmortem(request, fixturename):
    """Retrieve a fixture object if a test failed, else return None"""
    try:
        if request.node.report_setup.passed:
            try:
                if request.node.report_call.failed:
                    return request.node.funcargs.get(fixturename, None)
            # Handle case where `report_call` hasn't been generated yet
            # because the test hasn't finished running, as in the case of
            # a user interrupt
            except AttributeError:
                return None
    except AttributeError:
        return None


def pytest_collection_modifyitems(config, items):
    # add the index of each item in the list of items
    # this is used below for artifactory_result_path
    # to produce a unique result subdirectory for
    # each test (even if that test shares a name with
    # another test which is the case for parametrized tests).
    for i, item in enumerate(items):
        item.index = i


@pytest.fixture(scope="function", autouse=True)
def generate_artifactory_json(request, artifactory_repos):
    """Pytest fixture that leaves behind JSON upload and okify specfiles
    if the rtdata or rtdata_module fixtures are used in either a test or a
    module-scoped fixture that runs a pipeline and provides the results to a
    series of test.
    """
    inputs_root, results_root = artifactory_repos

    def artifactory_result_path():
        # Generate the Artifactory result path
        whoami = getpass.getuser() or "nobody"
        user_tag = f"NOT_CI_{whoami}"
        build_tag = os.environ.get("BUILD_TAG", user_tag)
        build_matrix_suffix = os.environ.get("BUILD_MATRIX_SUFFIX", "0")
        subdir = f"{TODAYS_DATE}_{build_tag}_{build_matrix_suffix}"
        testname = request.node.originalname or request.node.name
        basename = f"{request.node.index}_{testname}"

        return os.path.join(results_root, subdir, basename) + os.sep

    yield
    # Execute the following at test teardown
    upload_schema_pattern = []
    okify_schema_pattern = []

    rtdata = postmortem(request, "rtdata") or postmortem(request, "rtdata_module")
    if rtdata:
        try:
            # The function_jail fixture uses tmp_path
            cwd = str(request.node.funcargs["tmp_path"])
        except KeyError:
            # The module_jail fixture (module-scoped) returns the path
            cwd = str(request.node.funcargs["module_jail"])
        rtdata.remote_results_path = artifactory_result_path()
        rtdata.test_name = request.node.name
        # Dump the failed test traceback into rtdata
        rtdata.traceback = str(request.node.report_call.longrepr)

        # Upload and allow okify of truth by rtdata.output, if the test did not
        # fail before producing rtdata.output
        if rtdata.output and os.path.exists(rtdata.output):
            # Write the rtdata class out as an ASDF file
            path_asdf = os.path.join(cwd, f"{request.node.name}_rtdata.asdf")
            rtdata.to_asdf(path_asdf)

            # Generate an OKify JSON file
            pattern = os.path.join(
                rtdata.remote_results_path, os.path.basename(rtdata.output)
            )
            okify_schema_pattern.append(pattern)
            okify_schema = generate_upload_schema(
                okify_schema_pattern, f"{os.path.dirname(rtdata.truth_remote)}/"
            )

            jsonfile = os.path.join(cwd, f"{request.node.name}_okify.json")
            with open(jsonfile, "w") as fd:
                json.dump(okify_schema, fd, indent=2)

            # Generate an upload JSON file, including the OKify, asdf file
            upload_schema_pattern.append(rtdata.output)
            upload_schema_pattern.append(os.path.abspath(jsonfile))
            upload_schema_pattern.append(path_asdf)
            upload_schema = generate_upload_schema(
                upload_schema_pattern, rtdata.remote_results_path
            )

            jsonfile = os.path.join(cwd, f"{request.node.name}_results.json")
            with open(jsonfile, "w") as fd:
                json.dump(upload_schema, fd, indent=2)


def generate_upload_schema(pattern, target, recursive=False):
    """
    Generate JSON schema for Artifactory upload specfile using JFROG.

    Docs can be found at
    https://www.jfrog.com/confluence/display/RTF/Using+File+Specs

    Parameters
    ----------
    pattern : str or list of strings
        Specifies the local file system path to test results which should be
        uploaded to Artifactory. You can specify multiple artifacts by using
        wildcards or a regular expression as designated by the regexp property.

    target : str
        Specifies the target path in Artifactory in the following format::

            [repository_name]/[repository_path]

    recursive : bool, optional
        Specify whether or not to identify files listed in sub-directories
        for uploading.  Default: `False`

    Returns
    -------
    upload_schema : dict
        Dictionary specifying the upload schema
    """
    recursive = repr(recursive).lower()

    if not isinstance(pattern, str):
        # Populate schema for this test's data
        upload_schema = {"files": []}

        for p in pattern:
            temp_schema = copy.deepcopy(UPLOAD_SCHEMA["files"][0])
            temp_schema.update({"pattern": p, "target": target, "recursive": recursive})
            upload_schema["files"].append(temp_schema)
    else:
        # Populate schema for this test's data
        upload_schema = copy.deepcopy(UPLOAD_SCHEMA)
        upload_schema["files"][0].update(
            {"pattern": pattern, "target": target, "recursive": recursive}
        )
    return upload_schema


def _rtdata_fixture_implementation(artifactory_repos, envopt, request):
    """Provides the RemoteResource class"""
    inputs_root, results_root = artifactory_repos
    rtdata = RegtestData(env=envopt, inputs_root=inputs_root, results_root=results_root)

    yield rtdata


@pytest.fixture(scope="function")
def rtdata(artifactory_repos, envopt, request, function_jail):
    yield from _rtdata_fixture_implementation(artifactory_repos, envopt, request)


@pytest.fixture(scope="module")
def rtdata_module(artifactory_repos, envopt, request, module_jail):
    yield from _rtdata_fixture_implementation(artifactory_repos, envopt, request)


@pytest.fixture
def ignore_asdf_paths():
    ignore_attr = [
        # generic asdf stuff that will contain program version numbers
        # and other things that will almost certainly change in every case
        "asdf_library",
        "history",
        # roman-specific stuff to ignore
        "roman.meta.ref_file.crds.version",
        "roman.meta.calibration_software_version",
        "roman.cal_logs",
        "roman.meta.cal_logs",
        "roman.meta.date",
        "roman.meta.file_date",
        "roman.individual_image_cal_logs",
        "roman.meta.individual_image_meta",
    ]

    return {"ignore": ignore_attr}


@pytest.fixture(scope="module")
def resource_tracker():
    """Fixture to return the current module-scoped ResourceTracker.

    Use by calling ``track`` to generate a context in which resource
    usage will be tracked.

    >>>
    >> with resource_tracker.track():
    >>     # do stuff

    For resources used during tests providing a function-scoped
    request fixture result as the log argument will also log the
    used resources to the junit results.xml.

    >>>
    >> def test_something(resource_tracker, request):
    >>     with resource_tracker.track(log=request):
    >>         # do stuff

    For resources used during fixtures the tracked resources
    can be logged in a separate test using ``log_tracked_resources``.
    """
    return ResourceTracker()


@pytest.fixture()
def log_tracked_resources(resource_tracker, request):
    """Fixture to log resources tracked by ``resource_tracker``.

    >>>
    >> @pytest.fixture
    >> def my_fixture(resource_tracker):
    >>     with resource_tracker.track():
    >>         # do stuff
    >>
    >> def test_write_log(log_tracked_resources, my_fixture):
    >>     log_tracked_resources()
    """

    def callback():
        resource_tracker.log(request)

    yield callback



"""
Access the Roman Engineering Mnemonic Database.

The engineering mnemonics are provided by multiple services,
all of which require a level of authentication.

For non-operational use, the providing service is through the MAST AUI website
(TBD).

Authorization can be requested through https://auth.mast.stsci.edu/ site.

**Interface**

The primary entry point is the function `romancal.lib.engdb_tools.engdb_service`.
This function returns a `romancal.lib.engdb_lib.EngdbABC` connection object. Using
this object, values for a mnemonic covering a specified time range can be
retrieved using the :meth:`~romancal.lib.engdb_lib.EngdbABC.get_values` method.

By default, only values inclusively between the time end points are returned.
Depending on the frequency a mnemonic is updated, there can be no values. If
values are always desired, the nearest, bracketing values outside the time
range can be requested.

List of available services can be retrieved using `engdb_tools.AVAILABLE_SERVICES`

The primary service is the "mast" service. See `engdb_mast` for details.

.. warning::

    Many mnemonics are updated very quickly, up to 16Hz. When in doubt, specify a
    very short time frame, and request bracketing values. Otherwise, the request
    can return a very large amount of data, risking timeout, unnecessary memory
    consumption, or access restrictions.

Examples
--------
The typical workflow is as follows:

.. code-block:: python

    from romancal.lib.engdb.engdb_tools import engdb_service

    service = engdb_service()  # By default, will use the public MAST service.

    values = service.get_values("sa_zattest2", "2021-05-22T00:00:00", "2021-05-22T00:00:01")
"""

import logging

from .engdb_edp import EngdbEDP
from .engdb_mast import EngdbMast

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Define the available services
AVAILABLE_SERVICES = {
    "edp": EngdbEDP,
    "mast": EngdbMast,
}

# Expected errors from service initialization
EXPECTED_ERRORS = (KeyError, RuntimeError, TypeError)

__all__ = ["engdb_service"]


def engdb_service(service=None, **service_kwargs):
    """
    Provide access to the Roman Engineering Database.

    Access can be either through the public MAST API
    or by direct connection to the database server.

    Parameters
    ----------
    service : str or None
        The specific service to use. If None, first working service will be used.

    **service_kwargs : dict
        Service-specific keyword arguments. Refer to the concrete implementations
        of `~romancal.lib.engdb_lib.EngdbABC`.

    Returns
    -------
    engdb : `~romancal.lib.engdb_lib.EngdbABC`
        The engineering database service to use.
    """
    if service:
        try:
            engdb = AVAILABLE_SERVICES[service](**service_kwargs)
        except EXPECTED_ERRORS as excp:
            raise RuntimeError(f"Service {service} instantiation failed") from excp
        return engdb

    # No service was specified. Try until one is found.
    for name, service_class in AVAILABLE_SERVICES.items():
        try:
            engdb = service_class(**service_kwargs)
        except EXPECTED_ERRORS as excp:
            logger.debug("Service %s is unavailable.", name)
            logger.debug("Exception: %s", excp)
        else:
            # Found the working service. Continue on.
            break
    else:
        raise RuntimeError("No active engineering database service can be found.")

    # Service is in hand.
    return engdb



from pathlib import Path

from romancal.skycell import skymap

LAST_PROJREGION_INDEX = 1
DATA_DIRECTORY = Path(__file__).parent / "data"

if __name__ == "__main__":
    skymap_subset = skymap.SKYMAP.model.copy()

    # to maintain the proper indices, the subset must contain all the
    # previous projection regions up to the specified index
    skymap_subset.projection_regions = skymap.SKYMAP.model.projection_regions[
        : LAST_PROJREGION_INDEX + 1
    ].copy()

    skymap_subset.skycells = skymap.SKYMAP.model.skycells[
        : skymap_subset.projection_regions[-1]["skycell_end"] + 1
    ].copy()
    skymap_subset.save(DATA_DIRECTORY / "skymap_subset.asdf")



"""Test that TVAC/FPS run in the pipeline"""

from pathlib import Path

import pytest
import roman_datamodels as rdm
from gwcs.wcstools import grid_from_bounding_box

from romancal.lib.suffix import replace_suffix
from romancal.pipeline.exposure_pipeline import ExposurePipeline

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = pytest.mark.bigdata


@pytest.fixture(scope="module")
def run_elp(rtdata_module):
    rtdata = rtdata_module

    # Get reference
    rtdata.get_data("references/dark_ma510.asdf")

    input_data = "TVAC2_NOMOPS_WFIFLA_20240419194120_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "TVAC2_NOMOPS_WFIFLA_20240419194120_WFI01_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
        "--steps.tweakreg.skip=true",
        "--steps.source_catalog.skip=true",
        "--steps.dq_init.save=true",
        "--steps.dark_current.override_dark=dark_ma510.asdf",
    ]
    ExposurePipeline.from_cmdline(args)

    # get truth file
    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_elp):
    return run_elp.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_elp):
    return run_elp.truth


def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    ignore_asdf_paths["ignore"].append("roman.meta.ref_file.dark")
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_dqinit_matches_truth(run_elp, ignore_asdf_paths):
    output_path = Path(run_elp.output)
    dqinit_filename = replace_suffix(output_path.stem, "dqinit") + output_path.suffix
    dqinit_path = output_path.parent / dqinit_filename

    run_elp.output = dqinit_filename
    run_elp.get_truth(f"truth/WFI/image/{dqinit_filename}")

    truth_path = Path(run_elp.truth)

    diff = compare_asdf(dqinit_path, truth_path, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_output_is_image_model(output_model):
    assert isinstance(output_model, rdm.datamodels.ImageModel)


@pytest.mark.parametrize(
    "step_name",
    (
        "assign_wcs",
        "flat_field",
        "dark",
        "dq_init",
        "linearity",
        "ramp_fit",
        "saturation",
    ),
)
def test_steps_ran(output_model, step_name):
    assert getattr(output_model.meta.cal_step, step_name) == "COMPLETE"


def test_has_a_wcs(output_model):
    assert output_model.meta.wcs is not None


@pytest.mark.soctests
def test_wcs_has_distortion_information(output_model):
    assert "v2v3" in output_model.meta.wcs.available_frames


def test_wcs_applies_distortion_correction(output_model):
    # compare coordinates before and after distortion correction has been applied
    # 1 - get new image array based on the model
    x0, y0 = grid_from_bounding_box(output_model.meta.wcs.bounding_box)
    # 2 - apply the distortion-corrected WCS solution to new image array
    corrected_coords = output_model.meta.wcs(x0, y0)
    # 3 - apply the transformation from 'v2v3' to 'world' without distortion correction
    original_coords = output_model.meta.wcs.get_transform("v2v3", "world")(x0, y0)
    # compare both results to make sure they don't match
    # (which means the distortion correction was actually applied to the model)
    assert (corrected_coords[0] != original_coords[0]).all()
    assert (corrected_coords[1] != original_coords[1]).all()


@pytest.mark.parametrize("arr_name", ("dq", "err", "var_poisson", "var_rnoise"))
def test_array_exists(output_model, arr_name):
    assert hasattr(output_model, arr_name)


def test_has_exposure_time(output_model):
    assert "exposure_time" in output_model.meta.exposure


@pytest.mark.parametrize("meta_attribute", ("detector", "optical_element"))
def test_instrument_meta(output_model, meta_attribute):
    assert meta_attribute in output_model.meta.instrument


def test_wcs_has_bounding_box(output_model):
    assert len(output_model.meta.wcs.bounding_box) == 2



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from roman_datamodels import datamodels as rdm

from romancal.stpipe import RomanStep

if TYPE_CHECKING:
    from typing import ClassVar

__all__ = ["DarkCurrentStep"]

log = logging.getLogger(__name__)


class DarkCurrentStep(RomanStep):
    """DarkCurrentStep: Performs dark current correction by subtracting
    dark current reference data from the input science data model.
    """

    class_alias = "dark"

    spec = """
        dark_output = output_file(default = None) # Dark corrected model
    """

    reference_file_types: ClassVar = ["dark"]

    def process(self, input):
        if isinstance(input, rdm.DataModel):
            input_model = input
        else:
            # Open the input data model
            input_model = rdm.open(input)

        # Get the name of the dark reference file to use
        self.dark_name = self.get_reference_file(input_model, "dark")
        # Check for a valid reference file
        if self.dark_name == "N/A":
            log.warning("No DARK reference file found")
            log.warning("Dark current step will be skipped")
            result = input_model
            result.meta.cal_step.dark = "SKIPPED"
            return result

        log.info("Using DARK reference file: %s", self.dark_name)

        # Open dark model
        with rdm.open(self.dark_name) as dark_model:
            # get the dark slope and dark slope error from the reference file & trim ref pixels
            dark_slope = dark_model.dark_slope[4:-4, 4:-4]
            dark_slope_err = dark_model.dark_slope_error[4:-4, 4:-4]

            # Do the dark correction
            out_model = input_model
            out_model.data -= dark_slope
            out_model.err = np.sqrt(out_model.err**2 + (dark_slope_err) ** 2)
            out_model.dq |= dark_model.dq[4:-4, 4:-4]
            out_model.meta.cal_step.dark = "COMPLETE"

            # Save dark data to file
            if self.dark_output is not None:
                dark_model.save(self.dark_output)
                # not clear to me that this makes any sense for Roman

        if self.save_results:
            try:
                self.suffix = "darkcurrent"
            except AttributeError:
                self["suffix"] = "darkcurrent"

        return out_model



from pathlib import Path

import pytest
import roman_datamodels as rdm
from astropy.time import Time

from romancal.lib.suffix import replace_suffix
from romancal.step import FlatFieldStep
from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = [pytest.mark.bigdata]

IMAGE_FILENAME = "r0000101001001001001_0001_wfi01_f158_assignwcs.asdf"
GRISM_FILENAME = "r0000201001001001001_0001_wfi01_grism_assignwcs.asdf"


@pytest.fixture(
    params=[
        IMAGE_FILENAME,
        GRISM_FILENAME,
    ],
    scope="module",
)
def run_flat_field(rtdata_module, resource_tracker, request):
    rtdata = rtdata_module
    input_fn = request.param
    output_fn = replace_suffix(Path(input_fn).stem, "flat") + ".asdf"
    file_type = "grism" if "grism" in input_fn else "image"
    rtdata.get_data(f"WFI/{file_type}/{input_fn}")
    rtdata.output = output_fn
    rtdata.get_truth(f"truth/WFI/{file_type}/{output_fn}")
    with resource_tracker.track():
        RomanStep.from_cmdline(["romancal.step.FlatFieldStep", rtdata.input])
    return rtdata


@pytest.fixture(scope="module")
def output_model(run_flat_field):
    with rdm.open(run_flat_field.output) as model:
        yield model


def test_log_tracked_resources(log_tracked_resources, run_flat_field):
    log_tracked_resources()


def test_output_matches_truth(run_flat_field, ignore_asdf_paths):
    rtdata = run_flat_field
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_cal_step_updated(output_model):
    """Test that cal_step is step as expected"""
    if output_model.meta.exposure.type == "WFI_IMAGE":
        assert output_model.meta.cal_step.flat_field == "COMPLETE"
    else:
        assert output_model.meta.cal_step.flat_field == "SKIPPED"


def test_ref_file_used(output_model):
    """Test that a ref file was used and recorded where expected"""
    if output_model.meta.exposure.type == "WFI_IMAGE":
        assert "roman_wfi_flat" in output_model.meta.ref_file.flat
    else:
        assert output_model.meta.ref_file.flat == "N/A"


def test_ref_file_query(rtdata):
    """Test DMS79: 2 different start times return different flats"""
    rtdata.get_data(f"WFI/image/{IMAGE_FILENAME}")
    step = FlatFieldStep()
    with rdm.open(IMAGE_FILENAME) as model:
        ref_a = step.get_reference_file(model, "flat")
        model.meta.exposure.start_time = Time("2020-01-01T00:00:00", format="isot")
        ref_b = step.get_reference_file(model, "flat")
    assert ref_a != "N/A"
    assert ref_b != "N/A"
    assert ref_a != ref_b



"""Test module v1_calculate"""

from pathlib import Path

import numpy as np
import pytest
from astropy.table import Table
from astropy.time import Time
from roman_datamodels.datamodels import ScienceRawModel

import romancal.orientation.v1_calculate as v1c

DATA_PATH = Path(__file__).parent / "data"

# Engineering parameters
MAST_GOOD_STARTTIME = Time("2027-03-23T19:20:40", format="isot")
MAST_GOOD_ENDTIME = Time("2027-03-23T19:21:36", format="isot")


def test_from_models_mast(tmp_path):
    """Test v1_calculate_from_models for basic running"""
    model = ScienceRawModel.create_fake_data(
        {
            "meta": {
                "exposure": {
                    "start_time": MAST_GOOD_STARTTIME,
                    "end_time": MAST_GOOD_ENDTIME,
                },
            }
        }
    )

    try:
        v1_table = v1c.v1_calculate_from_models([model])
    except ValueError as exception:
        pytest.skip(
            f"MAST engineering database not available, possibly no token specified: {exception}"
        )
    v1_formatted = v1c.simplify_table(v1_table)

    # Save for post-test examination and read in to remove object infomation
    del v1_formatted["source"]
    v1_formatted.write(tmp_path / "test_from_models_mast.ecsv", format="ascii.ecsv")

    truth = Table.read(DATA_PATH / "test_from_models_mast.ecsv")
    errors = v1_compare_simplified_tables(v1_formatted, truth)
    errors_str = "\n".join(errors)
    assert len(errors) == 0, f"V1 tables are different: {errors_str}"


def test_over_time_mast(tmp_path):
    """Test v1_calculate_over_time for basic running"""
    try:
        v1_table = v1c.v1_calculate_over_time(MAST_GOOD_STARTTIME, MAST_GOOD_ENDTIME)
    except ValueError as exception:
        pytest.skip(
            f"MAST engineering database not available, possibly no token specified: {exception}"
        )
    v1_formatted = v1c.simplify_table(v1_table)

    # Save for post-test examination
    v1_formatted.write(tmp_path / "test_over_time_mast.ecsv", format="ascii.ecsv")

    truth = Table.read(DATA_PATH / "test_over_time_mast.ecsv")
    errors = v1_compare_simplified_tables(v1_formatted, truth)
    errors_str = "\n".join(errors)
    assert len(errors) == 0, f"V1 tables are different: {errors_str}"


# ######################
# Fixtures and utilities
# ######################
def v1_compare_simplified_tables(a, b, rtol=1e-05):
    """Calculate diff between tables generated by v1_calculate"""
    errors = []
    if len(a) != len(b):
        errors.append(f"len(a)={len(a)} not equal to len(b)={len(b)}")

    if not all(a["obstime"] == b["obstime"]):
        errors.append("obstimes different")

    if not np.allclose(a["ra"], b["ra"], rtol=rtol):
        errors.append("RAs are different")

    if not np.allclose(a["dec"], b["dec"], rtol=rtol):
        errors.append("DECs are different")

    if not np.allclose(a["pa_v3"], b["pa_v3"], rtol=rtol):
        errors.append("PA_V3s are different")

    return errors



"""Create an association from a list"""

import argparse
import sys

from .lib.rules_elpp_base import DMS_ELPP_Base
from .registry import AssociationRegistry

__all__ = ["asn_from_list"]


def asn_from_list(items, rule=DMS_ELPP_Base, **kwargs):
    """Create an association from a list

    Parameters
    ----------
    items: [object [, ...]]
        List of items to add.

    rule: `Association` rule
        The association rule to use.

    kwargs: dict
        Other named parameters required or pertinent to adding
        the items to the association.

    Returns
    -------
    association: `Association`-based instance
        The association with the items added.

    Notes
    -----
    This is a lower-level tool for artificially creating
    an association. As such, the association created may not be valid.
    It is presume the user knows what they are doing.
    """
    asn = rule()
    asn._add_items(items, **kwargs)
    if "target" in kwargs.keys():
        target = kwargs["target"]
        asn["target"] = target
    return asn


def _cli(args=None):
    """Command-line interface for list_to_asn

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """

    if args is None:
        args = sys.argv[1:]
    if isinstance(args, str):
        args = args.split(" ")

    parser = argparse.ArgumentParser(
        description="Create an association from a list of files",
        usage="asn_from_list -o mosaic_asn.json\n--product-name my_mosaic *.asdf",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        required=True,
        help="File to write association to",
    )

    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="json",
        help='Format of the association files. Default: "%(default)s"',
    )

    parser.add_argument(
        "--product-name",
        type=str,
        help="The product name when creating a Level 3 association",
    )

    parser.add_argument(
        "-r",
        "--rule",
        type=str,
        default="DMS_ELPP_Base",
        help=('The rule to base the association structure on. Default: "%(default)s"'),
    )
    parser.add_argument(
        "--ruledefs",
        action="append",
        help=(
            "Association rules definition file(s) If not specified, the default"
            " rules will be searched."
        ),
    )
    parser.add_argument(
        "-i",
        "--id",
        type=str,
        default="o999",
        help='The association candidate id to use. Default: "%(default)s"',
        dest="acid",
    )
    parser.add_argument(
        "-t",
        "--target",
        type=str,
        default="None",
        help='The target name for the association. Default: "%(default)s"',
        dest="target",
    )

    parser.add_argument(
        "filelist",
        type=str,
        nargs="+",
        help="File list to include in the association",
    )

    parsed = parser.parse_args(args=args)
    print("Parsed args:", parsed)

    # Get the rule
    rule = AssociationRegistry(parsed.ruledefs, include_bases=True)[parsed.rule]

    with open(parsed.output_file, "w") as outfile:
        asn = asn_from_list(
            parsed.filelist,
            rule=rule,
            product_name=parsed.product_name,
            acid=parsed.acid,
            target=parsed.target,
        )
        _, serialized = asn.dump(format=parsed.format)
        outfile.write(serialized)



"""
Pointing Summary.

Review contents of a set of given models for pointing information.
Compare the calculated V1 and REFPOINT pointing with the proposed
TARGET pointing.

Examples
--------
>>> from roman_datamodels.datamodels import ScienceRawModel
>>> im = ScienceRawModel.create_fake_data()
>>> im.meta.pointing.target_ra = 90.75541666666666
>>> im.meta.pointing.target_dec = -66.56055555555554
>>> im.meta.pointing.ra_v1 = 91.08142004561715
>>> im.meta.pointing.dec_v1 = -66.60547868904696
>>> im.meta.wcsinfo.ra_ref = 90.70377653291781
>>> im.meta.wcsinfo.dec_ref = -66.59540223936895
>>> calc_pointing_deltas(im)
Delta(target=<SkyCoord (ICRS): (ra, dec) in deg
    (90.75541667, -66.56055556)>, v1=<SkyCoord (ICRS): (ra, dec) in deg
    (91.08142005, -66.60547869)>, refpoint=<SkyCoord (ICRS): (ra, dec) in deg
    (90.70377653, -66.59540224)>, delta_v1=<Angle 0.13712727 deg>, delta_refpoint=<Angle 0.04044315 deg>)

>>> t = calc_deltas([im])
>>> print(t.columns)
<TableColumns names=('exposure','target','v1','refpoint','delta_v1','delta_refpoint')>
>>> delta_v1 = t["delta_v1"][0]
>>> "%.13f" % delta_v1
'0.1371272716401'
>>> delta_refpoint = t["delta_refpoint"][0]
>>> "%.13f" % delta_refpoint
'0.0404431476150'
"""

import logging
from collections import defaultdict, namedtuple

import astropy.units as u
import roman_datamodels as rdm
from astropy.coordinates import SkyCoord
from astropy.table import Table

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = ["Delta", "calc_deltas", "calc_pointing_deltas"]

# Basic delta structure
Delta = namedtuple("Delta", "target, v1, refpoint, delta_v1, delta_refpoint")


def calc_pointing_deltas(model):
    """
    Calculate pointing deltas.

    Parameters
    ----------
    model : `roman_datamodel.datamodels.RomanDataModel`
        The model to check pointing information in.

    Returns
    -------
    delta : Delta
        A `namedtuple` with the following keys. If for some reason
        a value could not be determined, `None` is given.

        - 'target': `SkyCoord` of the target.
        - 'v1': `SkyCoord` of V1.
        - 'refpoint': `SkyCoord` of the reference point.
        - 'delta_v1': Difference between V1 and proposed TARGET.
        - 'delta_refpoint': Difference between reference pixel pointing and TARGET.
    """
    # Retrieve the info from the model
    target = SkyCoord(
        model.meta.pointing.target_ra * u.degree,
        model.meta.pointing.target_dec * u.degree,
    )
    v1 = SkyCoord(
        model.meta.pointing.ra_v1 * u.degree, model.meta.pointing.dec_v1 * u.degree
    )
    refpoint = SkyCoord(
        model.meta.wcsinfo.ra_ref * u.degree, model.meta.wcsinfo.dec_ref * u.degree
    )

    # Calculate separations
    delta = Delta(
        target=target,
        v1=v1,
        refpoint=refpoint,
        delta_v1=target.separation(v1),
        delta_refpoint=target.separation(refpoint),
    )
    return delta


def calc_deltas(exposures, extra_meta=None):
    """
    Create table of pointing deltas.

    Parameters
    ----------
    exposures : [file-like[,...]] or [DataModel[,...]]
        List of file-like objects or `roman_datamodel.datamodels.RomanDataModel` to retrieve
        pointing information from.

    extra_meta : [str[,...]] or None
        List of model meta attributes to add to the table.

    Returns
    -------
    deltas : `astropy.table.Table`
        Table of results with the following columns:

        - exposure: The exposure the pointing information is from.
        - target: `SkyCoord` of the proposed target.
        - v1: `SkyCoord` of v1.
        - refpoint: `SkyCoord` of the reference point.
        - delta_v1: target - V1 separation.
        - delta_refpoint: target - refpoint separation.
    """
    # Initialize structures
    targets = []
    v1s = []
    refpoints = []
    delta_v1s = []
    delta_refpoints = []
    extra_meta_values = defaultdict(list)

    extra_meta = extra_meta if extra_meta is not None else []

    # Calculate deltas for all input.
    for exposure in exposures:
        with rdm.open(exposure) as model:
            delta = calc_pointing_deltas(model)
            logger.info(
                "%s: delta v1=%s delta refpoint=%s",
                model,
                delta.delta_v1,
                delta.delta_refpoint,
            )

            targets.append(delta.target)
            v1s.append(delta.v1)
            refpoints.append(delta.refpoint)
            delta_v1s.append(delta.delta_v1.degree)
            delta_refpoints.append(delta.delta_refpoint.degree)

            for meta in extra_meta:
                extra_meta_values[meta].append(model[meta])

    # Places results into a Table.
    deltas_dict = {
        "exposure": exposures,
        "target": targets,
        "v1": v1s,
        "refpoint": refpoints,
        "delta_v1": delta_v1s,
        "delta_refpoint": delta_refpoints,
    }
    deltas_dict.update(extra_meta_values)
    deltas = Table(deltas_dict)
    return deltas



"""Test utilities"""

import numpy as np
from astropy.stats import mad_std


def comp_wcs_grids_arcs(wcs_a, wcs_b, npix=4088, interval=10):
    """Compare world grids produced by the two wcs

    Parameters
    ----------
    wcs_a, wcs_b : gwcs.WCS
        The wcs object to compare.

    npix : int
        The size of the grid to produce.

    interval : int
        The interval to check over.

    Returns
    -------
    mad_std : float
        The numpy mad_std in mas
    """
    xx, yy = np.meshgrid(np.linspace(0, npix, interval), np.linspace(0, npix, interval))
    ra_a, dec_a = wcs_a(xx, yy, with_bounding_box=False)
    ra_b, dec_b = wcs_b(xx, yy, with_bounding_box=False)
    dec_med = np.nanmedian(dec_b)

    ra_mad = (
        mad_std((ra_a - ra_b) * np.cos(np.radians(dec_med)), ignore_nan=True)
        * 60.0
        * 60.0
        * 1000.0
    )
    dec_mad = mad_std(dec_a - dec_b, ignore_nan=True) * 60.0 * 60.0 * 1000.0

    return ra_mad, dec_mad



"""
Module to calculate PSF photometry.
"""

import logging
import math
from collections import OrderedDict

import astropy.units as u
import numpy as np
from astropy.convolution import Box2DKernel, convolve
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.nddata import NDData
from astropy.table import Table
from astropy.utils import lazyproperty
from photutils.background import LocalBackground
from photutils.detection import DAOStarFinder
from photutils.psf import (
    GriddedPSFModel,
    ImagePSF,
    IterativePSFPhotometry,
    PSFPhotometry,
    SourceGrouper,
)
from scipy.ndimage import map_coordinates

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def azimuthally_smooth(data, oversample=2, scaling=1.0, order=4):
    """Azimuthally smooth model

    The image is converted to polar coordinates via a 4th order spline interpolation.
    The image average is determined at each radius, and a final image is constructed by reprojecting this averaged
    image back into cartesian coordinates.

    Parameters
    ----------
    data : nd.array(size=(*, *))
        Data to be smoothed.

    oversample : int
        Oversampling of the data to improve fidelity of the conversions
        between cartesian and polar layout

    scaling : float
        Scale factor to apply to the result.

    order : int
        Order of the spline interpolation used for the coordinate transformations

    Returns
    -------
    smoothed : nd.array(size=(*, *))
        The azimuthally smoothed image
    """

    # Define cartesian->polar->cartesian conversion.
    # Significant difference is the polar back to cartesian uses
    # the original data's transformation parameters, to reproduce
    # an image of the correct dimensions.
    def cart_to_polar(model, oversample=2, order=4):
        ntheta = 4 * model.shape[0] * oversample
        nrad = model.shape[0] * oversample
        szo2 = model.shape[0] // 2
        rr = np.linspace(-5, szo2 * np.sqrt(2), nrad)
        tt = np.linspace(0, 2 * np.pi, ntheta, endpoint=False)
        xx = rr[:, None] * np.cos(tt[None, :]) + szo2
        yy = rr[:, None] * np.sin(tt[None, :]) + szo2
        modelpolar = map_coordinates(
            model, [xx, yy], order=order, mode="nearest", output=np.dtype("f4")
        )
        return modelpolar, rr, tt

    def polar_to_cart(model, rr, tt, scaling=1.0, order=4):
        szo = model.shape[0]
        szo = math.ceil(szo / scaling)
        szo2 = szo // 2
        xo, yo = np.mgrid[-szo2 : szo2 + 1, -szo2 : szo2 + 1] * scaling
        ro = np.sqrt(xo**2 + yo**2)
        to = np.arctan2(yo, xo) % (2 * np.pi)
        ro = np.interp(ro, rr, np.arange(len(rr)).astype("f4"))
        to = np.interp(to, tt, np.arange(len(tt)).astype("f4"))
        res = map_coordinates(
            model, [ro, to], order=order, mode="wrap", output=np.dtype("f4")
        )
        res *= scaling**2  # keep normalization at 1
        return res

    polar, rr, tt = cart_to_polar(data, oversample=oversample)
    polar_mean = np.mean(polar, axis=1, keepdims=True)
    smoothed = polar_to_cart(polar_mean, rr, tt, scaling=scaling)

    return smoothed


def get_gridded_psf_model(psf_ref_model):
    """Function to generate gridded PSF model from psf reference file

    Compute a gridded PSF model for one SCA using the
    reference files in CRDS.
    The input reference files have 3 focus positions and this is using
    the in-focus images. There are also three spectral types that are
    available and this code uses the M5V spectal type.
    """
    # Open the reference file data model
    # select the infocus images (0) and we have a selection of spectral types
    # A0V, G2V, and M6V, pick G2V (1)
    focus = 0
    spectral_type = 1
    psf_images = psf_ref_model.psf[focus, spectral_type, :, :, :].copy()
    # get the central position of the cutouts in a list
    psf_positions_x = psf_ref_model.meta.pixel_x.data.data
    psf_positions_y = psf_ref_model.meta.pixel_y.data.data
    meta = OrderedDict()
    position_list = []
    for index in range(len(psf_positions_x)):
        position_list.append([psf_positions_x[index], psf_positions_y[index]])

    # integrate over the native pixel scale
    oversample = psf_ref_model.meta.oversample
    pixel_response_kernel = Box2DKernel(width=oversample)
    for i in range(psf_images.shape[0]):
        psf = psf_images[i, :, :]
        im = convolve(psf, pixel_response_kernel) * oversample**2
        psf_images[i, :, :] = im

    meta["grid_xypos"] = position_list
    meta["oversampling"] = oversample
    nd = NDData(psf_images, meta=meta)
    model = GriddedPSFModel(nd)

    return model


def create_l3_psf_model(
    psf_ref_model,
    pixfrac=1.0,
    pixel_scale=0.11,
    oversample=11,
):
    """
    Compute a PSF model for an L3 image.

    L3 data is an amalgamation of numerous exposures over numerous SCA's.
    This algorithm does not attempt to merge specific PSF profiles for each
    SCA that contributes to each pixel. Instead, a simplified version is implemented
    as described.

        - Base PSF for the given detector, is created via get_gridded_psf_model.
          This bas has the native 0.11 arcsec pixel scale already convolved in.
        - PSF is further convolved with the drizzlepac `pixfrac` scale
        - PSF is further convolved with the images actual pixel scale.
        - The PSF is then azimuthally averaged and resampled at the L3 pixel scale.

    Parameters
    ----------
    psf_ref_model : str
        PSF reference file model to use
    pixfrac : float
        drizzlepac pixel fraction used.
    pixel_scale : float
        L3 image pixel scale in arcsec.
        Often similar to the default detector scale of 0.11 arcsec.
    oversample : int, optional
        Oversample factor, default is 11.

    Returns
    -------
    psf_model : `photutils.psf.ImagePSF`
        PSF model.

    """

    gridpsf = get_gridded_psf_model(psf_ref_model)
    center = 2044  # Roman SCA center pixel
    szo2 = 10
    npix = szo2 * 2 * oversample + 1
    pts = np.linspace(-szo2 + center, szo2 + center, npix)
    xx, yy = np.meshgrid(pts, pts)
    psf = gridpsf.evaluate(xx, yy, 1, center, center)
    detector_pixel_scale = 0.11  # Roman native scale
    # psf is now a stamp going 10 pixels out from the center of the PSF
    # at the native PSF scale, oversampled by a factor of oversample.

    # Smooth to account for the pixfrac used to create the L3 image.
    pixfrac_kernel = Box2DKernel(width=pixfrac * oversample)
    psf = convolve(psf, kernel=pixfrac_kernel)

    # Smooth to the image scale
    outscale_kernel = Box2DKernel(width=oversample * pixel_scale / detector_pixel_scale)
    psf = convolve(psf, kernel=outscale_kernel)

    # Azimuthally smooth the psf
    psf = azimuthally_smooth(psf, scaling=pixel_scale / detector_pixel_scale)

    # Create the PSF model.
    x_0, y_0 = psf.shape
    x_0 = (x_0 - 1) / 2.0 / oversample
    y_0 = (y_0 - 1) / 2.0 / oversample
    psf_model = ImagePSF(psf, x_0=x_0, y_0=y_0, oversampling=oversample)

    return psf_model


def fit_psf_to_image_model(
    image_model=None,
    data=None,
    error=None,
    mask=None,
    photometry_cls=PSFPhotometry,
    psf_model=None,
    grouper=None,
    fitter=None,
    localbkg_estimator=None,
    finder=None,
    x_init=None,
    y_init=None,
    progress_bar=False,
    error_lower_limit=None,
    fit_shape=(15, 15),
    exclude_out_of_bounds=True,
):
    """
    Fit PSF models to an ``ImageModel``.

    Parameters
    ----------
    image_model : `roman_datamodels.datamodels.ImageModel`
        Image datamodel. If ``image_model`` is supplied,
        ``data,error`` should be `None`.
    data : `astropy.units.Quantity`
        Fit a PSF model to the rate image ``data``.
        If ``data,error`` are supplied, ``image_model`` should be `None`.
    error : `astropy.units.Quantity`
        Uncertainties on fluxes in ``data``. Should be `None` if
        ``image_model`` is supplied.
    mask : 2D bool `numpy.ndarray`, optional
        Mask to apply to the data. Default is `None`.
    photometry_cls : {`photutils.psf.PSFPhotometry`,
            `photutils.psf.IterativePSFPhotometry`}
        Choose a photutils PSF photometry technique (default or iterative).
    psf_model : `astropy.modeling.Fittable2DModel`
        The 2D PSF model to fit to the rate image. Usually this model is an instance
        of `photutils.psf.GriddedPSFModel`.
    grouper : `photutils.psf.SourceGrouper`
        Specifies rules for attempting joint fits of multiple PSFs when
         there are nearby sources at small separations.
    fitter : `astropy.modeling.fitting.Fitter`, optional
        Modeling class which optimizes the PSF fit.
        Default is `astropy.modeling.fitting.LevMarLSQFitter(calc_uncertainties=True)`.
    localbkg_estimator : `photutils.background.LocalBackground`, optional
        Specifies inner and outer radii for computing flux background near
        a source. Default has ``inner_radius=10, outer_radius=30``.
    finder : subclass of `photutils.detection.StarFinderBase`, optional
        When ``photutils_cls`` is `photutils.psf.IterativePSFPhotometry`, the
        ``finder`` is called to determine if sources remain in the rate image
        after one PSF model is fit to the observations and removed.
        Default was extracted from the `DAOStarFinder` call in the
        Source Detection step.
    x_init : `numpy.ndarray`, optional
        Initial guesses for the ``x`` pixel coordinates of each source to fit.
    y_init : `numpy.ndarray`, optional
        Initial guesses for the ``y`` pixel coordinates of each source to fit.
    progress_bar : bool, optional
        Render a progress bar via photutils. Default is False.
    error_lower_limit : `astropy.units.Quantity`, optional
        Since some synthetic images may have bright sources with very
        small statistical uncertainties, the ``error`` can be clipped at
        ``error_lower_limit`` to prevent over-confident fits.
    fit_shape : int, or tuple of length 2, optional
        Rectangular shape around the center of a star that will
        be used to define the PSF-fitting data. See docs for
        `photutils.psf.PSFPhotometry` for details. Default is ``(16, 16)``.
    exclude_out_of_bounds : bool, optional
        If `True`, do not attempt to fit stars which have initial centroids
        that fall outside the pixel limits of the SCA. Default is False.

    Returns
    -------
    results_table : `astropy.table.QTable`
        PSF photometry results.
    photometry : instance of class ``photutils_cls``
        PSF photometry instance with configuration settings and results.

    """
    if grouper is None:
        # minimum separation before sources are fit simultaneously:
        grouper = SourceGrouper(min_separation=5)  # [pix]

    if fitter is None:
        fitter = LevMarLSQFitter(calc_uncertainties=True)

    # the iterative PSF method requires a finder:
    psf_photometry_kwargs = {}
    if photometry_cls is IterativePSFPhotometry or (x_init is None and y_init is None):
        if finder is None:
            # these defaults extracted from the
            # romancal SourceDetectionStep
            finder = DAOStarFinder(
                fwhm=1.0,
                threshold=0.0,
                sharplo=0.0,
                sharphi=1.0,
                roundlo=-1.0,
                roundhi=1.0,
                peakmax=None,
            )

        psf_photometry_kwargs["finder"] = finder

    if localbkg_estimator is None:
        localbkg_estimator = LocalBackground(
            inner_radius=10,  # [pix]
            outer_radius=30,  # [pix]
        )

    photometry = photometry_cls(
        grouper=grouper,
        localbkg_estimator=localbkg_estimator,
        psf_model=psf_model,
        fitter=fitter,
        fit_shape=fit_shape,
        aperture_radius=fit_shape[0],
        progress_bar=progress_bar,
        **psf_photometry_kwargs,
    )

    if x_init is not None and y_init is not None:
        guesses = Table(np.column_stack([x_init, y_init]), names=["x_init", "y_init"])
    else:
        guesses = None

    if image_model is None:
        if data is None and error is None:
            raise ValueError(
                "PSF fitting requires either an ImageModel, "
                "or arrays for the data and error."
            )

    if data is None and image_model is not None:
        data = image_model.data

    if error is None and image_model is not None:
        error = image_model.err

    if error_lower_limit is not None:
        # option to enforce a lower limit on the flux uncertainties
        error = np.clip(error, error_lower_limit, None)

    if exclude_out_of_bounds and guesses is not None:
        # don't attempt to fit PSFs for objects with initial centroids
        # outside the detector boundaries:
        init_centroid_in_range = (
            (guesses["x_init"] > 0)
            & (guesses["x_init"] < data.shape[1])
            & (guesses["y_init"] > 0)
            & (guesses["y_init"] < data.shape[0])
        )
        guesses = guesses[init_centroid_in_range]

    # fit the model PSF to the data:
    results_table = photometry(data=data, error=error, init_params=guesses, mask=mask)

    # results are stored on the PSFPhotometry instance:
    return results_table, photometry


class PSFCatalog:
    """
    Class to calculate PSF photometry.

    Parameters
    ----------
    model : `ImageModel` or `MosaicModel`
        The input data model. The image data is assumed to be background
        subtracted.

    xypos : `numpy.ndarray`
        Pixel coordinates of sources to fit. The shape of this array should
        be (N, 2), where N is the number of sources to fit.

    mask : 2D `~numpy.ndarray` or `None`, optional
        A 2D boolean mask image with the same shape as the input data.
        This mask is used for PSF photometry. The mask should be the
        same one used to create the segmentation image.
    """

    def __init__(self, model, psf_ref_model, xypos, mask=None):
        self.model = model
        self.psf_ref_model = psf_ref_model
        self.xypos = xypos
        self.mask = mask

        self.names = list(self._name_map.values())
        self.names.extend(["ra_psf", "dec_psf", "ra_psf_err", "dec_psf_err"])

        self.calc_psf_photometry()

    @lazyproperty
    def psf_model(self):
        """
        A gridded PSF model based on instrument and detector
        information.
        """
        if hasattr(self.model.meta, "instrument"):
            # ImageModel (L2 datamodel)
            psf_model = get_gridded_psf_model(self.psf_ref_model)
        else:
            # MosaicModel (L3 datamodel)
            psf_model = create_l3_psf_model(
                self.psf_ref_model,
                pixel_scale=self.model.meta.wcsinfo.pixel_scale
                * 3600.0,  # wcsinfo is in degrees. Need arcsec
                pixfrac=self.model.meta.resample.pixfrac,
            )

        return psf_model

    @lazyproperty
    def _name_map(self):
        """
        Mapping of photutils column names to the output catalog names.
        """
        name_map = {}
        name_map["x_fit"] = "x_psf"
        name_map["x_err"] = "x_psf_err"
        name_map["y_fit"] = "y_psf"
        name_map["y_err"] = "y_psf_err"
        name_map["flux_fit"] = "psf_flux"
        name_map["flux_err"] = "psf_flux_err"
        name_map["qfit"] = "psf_gof"
        name_map["flags"] = "psf_flags"
        return name_map

    def calc_psf_photometry(self):
        """
        Perform PSF photometry by fitting PSF models to detected
        sources.
        """
        xinit, yinit = np.transpose(self.xypos)
        psf_photometry_table, _ = fit_psf_to_image_model(
            image_model=self.model,
            mask=self.mask,
            psf_model=self.psf_model,
            x_init=xinit,
            y_init=yinit,
            exclude_out_of_bounds=True,
        )

        # set these columns as attributes of this instance
        for old_name, new_name in self._name_map.items():
            value = psf_photometry_table[old_name]

            # change the photutils dtypes
            if np.issubdtype(value.dtype, np.integer):
                value = value.astype(np.int32)
            elif np.issubdtype(value.dtype, np.floating):
                value = value.astype(np.float32)

            # handle any unit conversions
            if new_name in ("x_psf", "y_psf", "x_psf_err", "y_psf_err"):
                value *= u.pix

            setattr(self, new_name, value)

    @lazyproperty
    def _sky_psf(self):
        """
        The pixel coordinates of the fitted-PSF pixel position.
        """
        return self.model.meta.wcs.pixel_to_world(self.x_psf, self.y_psf)

    @lazyproperty
    def ra_psf(self):
        """
        Right Ascension of the fitted-PSF pixel position.
        """
        return self._sky_psf.ra

    @lazyproperty
    def dec_psf(self):
        """
        Declination of the fitted-PSF pixel position.
        """
        return self._sky_psf.dec

    @lazyproperty
    def ra_psf_err(self):
        """
        The uncertainty in ra_psf.
        """
        return np.zeros(self.ra_psf.shape, dtype=np.float32) * u.arcsec

    @lazyproperty
    def dec_psf_err(self):
        """
        The uncertainty in dec_psf.
        """
        return np.zeros(self.dec_psf.shape, dtype=np.float32) * u.arcsec



"""
Unit tests for skycell.match.

These tests depend very strongly on the contents of the referenced table of patches.
Changes to the contents of this table will require changes to the tests for
any unit tests that depend on specific matches to patches in the table.

Any changes to the matching algorithm should be completely separate from any
changes to the table contents.

The tests include the following cases to validate the matches
1) Simple case at ra, dec, pa = 0, 0, 0
2) 1) translated to ra, de, pa = 180, 40, 0
3) Same as 2) but with pa = 45 and 60.
4) Same as 3) (pa=45 case) but with the lower corner just above, below, to the right
   and to the left of a 4 patch corner
   (assuming non-overlapping patchs within a common tangent point).
   This requires identifying the ra, dec of such a corner in the table.
5) A test of a WCS provided example.

Most of these tests check to see if the matches are what are expected, by index of
the table (the table format is expected to include the index of the entry as one
of its columns so that subsets of the table selected to reduce the filesize still
retain the same index obtained.)
"""

from pathlib import Path

import astropy.coordinates as coord
import astropy.modeling.models as amm
import astropy.units as u
import numpy as np
import pytest
import spherical_geometry.vector as sgv
from gwcs import WCS, coordinate_frames

import romancal.skycell.match as sm
from romancal.skycell import skymap

TEST_POINTS = [
    (0.88955854, 87.53857137),
    (20.6543883, 87.60498618),
    (343.19474696, 85.05565535),
    (8.94286202, 85.50465173),
    (27.38417684, 85.03404907),
    (310.53503934, 88.56749324),
]
EPSILON = 0.0011  # epsilon offset in degrees
DATA_DIRECTORY = Path(__file__).parent / "data"


@pytest.fixture()
def skymap_subset() -> skymap.SkyMap:
    """
    smaller subset to allow these tests
    to run without access to the full skymap from CRDS.
    """
    return skymap.SkyMap(DATA_DIRECTORY / "skymap_subset.asdf")


def mk_im_corners(
    ra: float, dec: float, pa: float, size: float
) -> list[tuple[float, float]]:
    """
    Generate 4 image corners of a square with the center at the supplied
    side size, ra, dec, and position angle (all in degrees).
    """
    # Generate 4 unit vectors at ra, dec = (0 , 0)
    center = sgv.lonlat_to_vector(0.0, 0.0)
    radecvec = sgv.lonlat_to_vector(ra, dec)
    zaxis = (0.0, 0.0, 1.0)
    yaxis = (0.0, 1.0, 0.0)
    pp = sgv.rotate_around(
        *(sgv.rotate_around(*(center + yaxis + (-size / 2,))) + zaxis + (+size / 2,))
    )
    pm = sgv.rotate_around(
        *(sgv.rotate_around(*(center + yaxis + (+size / 2,))) + zaxis + (+size / 2,))
    )
    mp = sgv.rotate_around(
        *(sgv.rotate_around(*(center + yaxis + (-size / 2,))) + zaxis + (-size / 2,))
    )
    mm = sgv.rotate_around(
        *(sgv.rotate_around(*(center + yaxis + (+size / 2,))) + zaxis + (-size / 2,))
    )
    rect = [pp, mp, mm, pm]

    # Now move to requested ra and dec
    trect = [
        sgv.rotate_around(
            *(sgv.rotate_around(*(vec + yaxis + (-dec,))) + zaxis + (ra,))
        )
        for vec in rect
    ]
    # Rotate to desired position angle
    rrect = [sgv.rotate_around(*(vec + radecvec + (pa,))) for vec in trect]
    frect = [sgv.vector_to_lonlat(*vec) for vec in rrect]
    # Reorganize by ra, dec arrays
    radecrect = np.array(frect)
    return radecrect


def mk_gwcs(ra, dec, pa, bounding_box=None) -> WCS:
    """
    Construct a GWCS model for testing the patch matching when provided a WCS
    This just implements a basic tangent projection with specified ra, dec, and
    position angle
    """
    transform = (amm.Shift(-2048) & amm.Shift(-2048)) | (
        amm.Scale(0.11 / 3600.0) & amm.Scale(0.11 / 3600.0)
        | amm.Rotation2D(pa)
        | amm.Pix2Sky_TAN()
        | amm.RotateNative2Celestial(ra, dec, 180.0)
    )
    detector_frame = coordinate_frames.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = coordinate_frames.CelestialFrame(
        reference_frame=coord.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )
    wcsobj = WCS([(detector_frame, transform), (sky_frame, None)])
    if bounding_box is not None:
        wcsobj.bounding_box = bounding_box
        wcsobj.array_shape = tuple(
            int(bounding_box[index][1] - bounding_box[index][0])
            for index in range(len(bounding_box))
        )
    return wcsobj


@pytest.mark.parametrize(
    "test_point,offset,rotation,size,expected_skycell_names",
    [
        (
            TEST_POINTS[0],
            (0, 0),
            45,
            0.001,
            (
                "000p86x66y51",
                "000p86x66y50",
                "000p86x65y51",
                "000p86x65y50",
            ),
        ),
        (
            TEST_POINTS[0],
            (0, +EPSILON),
            45,
            0.001,
            (
                "000p86x66y50",
                "000p86x66y51",
            ),
        ),
        (
            TEST_POINTS[1],
            (0, +EPSILON),
            45,
            0.001,
            (
                "000p86x69y61",
                "000p86x69y62",
            ),
        ),
        (
            TEST_POINTS[1],
            (0, -EPSILON),
            45,
            0.001,
            ("000p86x69y62", "000p86x68y62", "000p86x69y61", "000p86x68y61"),
        ),
        (
            TEST_POINTS[1],
            (+EPSILON, 0),
            45,
            0.001,
            ("000p86x69y62", "000p86x68y62", "000p86x69y61", "000p86x68y61"),
        ),
        (
            TEST_POINTS[1],
            (-EPSILON, 0),
            45,
            0.001,
            ("000p86x69y62", "000p86x69y61", "000p86x68y62", "000p86x68y61"),
        ),
        (
            TEST_POINTS[1],
            (0, 0),
            45,
            0.001,
            ("000p86x69y62", "000p86x68y62", "000p86x69y61", "000p86x68y61"),
        ),
        (
            TEST_POINTS[0],
            (0, 0),
            45,
            0.3,
            (
                "000p86x66y51",
                "000p86x66y50",
                "000p86x65y51",
                "000p86x65y50",
                "000p86x67y51",
                "000p86x66y52",
                "000p86x67y50",
                "000p86x65y52",
                "000p86x64y51",
                "000p86x66y49",
                "000p86x64y50",
                "000p86x65y49",
                "000p86x67y52",
                "000p86x67y49",
                "000p86x64y52",
                "000p86x64y49",
                "000p86x68y51",
                "000p86x66y53",
                "000p86x68y50",
                "000p86x65y53",
                "000p86x63y51",
                "000p86x66y48",
                "000p86x63y50",
                "000p86x65y48",
            ),
        ),
        (
            TEST_POINTS[1],
            (0, 0),
            45,
            0.5,
            (
                "000p86x69y62",
                "000p86x68y62",
                "000p86x69y61",
                "000p86x68y61",
                "000p86x70y62",
                "000p86x69y63",
                "000p86x70y61",
                "000p86x68y63",
                "000p86x67y62",
                "000p86x69y60",
                "000p86x67y61",
                "000p86x68y60",
                "000p86x70y60",
                "000p86x67y63",
                "000p86x67y60",
                "000p86x71y62",
                "000p86x71y61",
                "000p86x66y62",
                "000p86x69y59",
                "000p86x66y61",
                "000p86x68y59",
                "000p86x71y60",
                "000p86x66y63",
                "000p86x70y59",
                "000p86x66y60",
                "000p86x67y59",
                "000p86x72y61",
                "000p86x71y59",
                "000p86x66y64",
                "000p86x65y62",
                "000p86x69y58",
                "000p86x65y61",
                "000p86x68y58",
                "000p86x66y59",
                "000p86x72y60",
                "000p86x65y63",
                "000p86x70y58",
                "000p86x65y60",
                "000p86x67y58",
                "000p86x72y59",
                "000p86x65y64",
                "000p86x66y58",
                "000p86x73y61",
                "000p86x64y62",
                "000p86x68y57",
                "000p86x73y60",
                "000p86x64y63",
                "000p86x67y57",
            ),
        ),
        (
            TEST_POINTS[2],
            (0, 0),
            0,
            0.4,
            (
                "000p86x35y31",
                "000p86x35y30",
                "000p86x34y31",
                "000p86x34y30",
                "000p86x35y32",
                "000p86x36y31",
                "000p86x34y32",
                "000p86x36y30",
                "000p86x35y29",
                "000p86x33y31",
                "000p86x34y29",
                "000p86x33y30",
                "000p86x36y32",
                "000p86x36y29",
                "000p86x33y32",
                "000p86x33y29",
                "000p86x35y33",
                "000p86x37y31",
                "000p86x34y33",
                "000p86x37y30",
                "000p86x35y28",
                "000p86x32y31",
                "000p86x34y28",
                "000p86x32y30",
                "000p86x36y33",
                "000p86x37y32",
                "000p86x33y33",
                "000p86x37y29",
                "000p86x36y28",
                "000p86x32y32",
                "000p86x33y28",
                "000p86x32y29",
                "000p86x37y33",
                "000p86x35y34",
                "000p86x38y30",
                "000p86x37y28",
                "000p86x32y33",
                "000p86x31y31",
                "000p86x34y27",
                "000p86x32y28",
                "000p86x36y34",
                "000p86x38y29",
                "000p86x31y32",
                "000p86x33y27",
                "000p86x37y34",
            ),
        ),
        (
            TEST_POINTS[3],
            (-0.5, -0.5),
            0,
            0.001,
            ("000p86x32y60",),
        ),
        (
            TEST_POINTS[4],
            (0, 0),
            -62,
            0.2,
            (),
        ),
        (
            TEST_POINTS[5],
            (0, 0),
            188,
            0.25,
            (
                "135p90x69y52",
                "135p90x70y52",
                "135p90x69y51",
                "135p90x70y51",
                "135p90x68y52",
                "135p90x69y53",
                "135p90x68y51",
                "135p90x70y53",
                "135p90x71y52",
                "135p90x69y50",
                "135p90x71y51",
                "135p90x70y50",
                "135p90x68y53",
                "135p90x71y53",
                "135p90x68y50",
                "135p90x71y50",
            ),
        ),
    ],
)
def test_skycell_match(
    test_point, offset, rotation, size, expected_skycell_names, skymap_subset
):
    corners = mk_im_corners(*test_point + np.array(offset), rotation, size)

    intersecting_skycells = sm.find_skycell_matches(corners, skymap=skymap_subset)

    skycell_names = np.array(
        [skymap_subset.model.skycells[index]["name"] for index in intersecting_skycells]
    ).tolist()

    assert sorted(skycell_names) == sorted(expected_skycell_names)


@pytest.mark.parametrize(
    "test_point,expected_skycell_names",
    [
        (
            TEST_POINTS[1],
            [
                "000p86x69y62",
                "000p86x69y61",
                "000p86x68y62",
                "000p86x68y61",
                "000p86x69y63",
                "000p86x70y61",
                "000p86x67y62",
                "000p86x68y60",
            ],
        )
    ],
)
def test_match_from_wcs_with_bbox(test_point, expected_skycell_names, skymap_subset):
    wcsobj = mk_gwcs(
        *test_point,
        45,
        bounding_box=((-0.5, 4096 - 0.5), (-0.5, 4096 - 0.5)),
    )

    intersecting_skycells = sm.find_skycell_matches(wcsobj, skymap=skymap_subset)

    skycell_names = np.array(
        [skymap_subset.model.skycells[index]["name"] for index in intersecting_skycells]
    ).tolist()

    assert skycell_names == expected_skycell_names


@pytest.mark.parametrize("test_point", [TEST_POINTS[1]])
def test_match_from_wcs_without_bbox(test_point):
    wcsobj = mk_gwcs(*test_point, 45)

    with pytest.raises(ValueError):
        sm.find_skycell_matches(wcsobj, skymap=skymap_subset)



"""
Module to calculate the source catalog.
"""

import logging

import astropy.units as u
import numpy as np
from astropy.table import QTable, Table
from astropy.utils import lazyproperty
from roman_datamodels.datamodels import ImageModel, MosaicModel
from roman_datamodels.dqflags import pixel

from romancal import __version__ as romancal_version
from romancal.source_catalog.aperture import ApertureCatalog
from romancal.source_catalog.daofind import DAOFindCatalog
from romancal.source_catalog.neighbors import NNCatalog
from romancal.source_catalog.psf import PSFCatalog
from romancal.source_catalog.segment import SegmentCatalog

from ._wcs_helpers import pixel_scale_angle_at_skycoord

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RomanSourceCatalog:
    """
    Class for the Roman source catalog.

    Parameters
    ----------
    model : `ImageModel` or `MosaicModel`
        The input data model. The image data is assumed to be background
        subtracted.

    segment_image : `~photutils.segmentation.SegmentationImage`
        A 2D segmentation image, with the same shape as the input data,
        where sources are marked by different positive integer values. A
        value of zero is reserved for the background.

    convolved_data : 2D `~numpy.ndarray` or `None`
        The 2D array used to calculate the source centroid and shape
        measurements. The image is assumed to be background subtracted.

    kernel_fwhm : float
        The full-width at half-maximum (FWHM) of the DAOFind 2D Gaussian
        kernel. This kernel is used to calculate the DAOFind sharpness
        and roundness properties. DAOFind uses a special kernel that
        sums to zero.

    fit_psf : bool, optional
        Whether to fit a PSF model to the sources.

    mask : 2D `~numpy.ndarray` or `None`, optional
        A 2D boolean mask image with the same shape as the input data.
        This mask is used for PSF photometry. The mask should be the
        same one used to create the segmentation image.

    detection_cat : `None` or `RomanSourceCatalog`, optional
        A `RomanSourceCatalog` object for the detection image. The
        segmentation image used to create the detection catalog must
        be the same one input to ``segment_image``. If input, then the
        detection catalog source centroids and morphological/shape
        properties will be returned instead of calculating them from
        the input ``data``. The detection catalog centroids and shape
        properties will also be used to perform aperture photometry
        (i.e., circular and Kron).

    flux_unit : str, optional
        The unit of the flux density. Default is 'nJy'.

    cat_type : {'prompt', 'dr_det', 'dr_band', 'forced_full', 'forced_det'}, optional
        The type of catalog to create. The default is 'prompt'. This
        determines the columns in the output catalog. The 'dr_det' and
        'dr_band' catalogs are band-specific catalogs for the
        multiband source detection.

    Notes
    -----
    ``model.err`` is assumed to be the total error array corresponding
    to the input science ``model.data`` array. It is assumed to include
    all sources of error, including the Poisson error of the sources. It
    must also have the same shape and units as the science data array.
    """

    def __init__(
        self,
        model,
        segment_img,
        convolved_data,
        kernel_fwhm,
        *,
        fit_psf=True,
        mask=None,
        psf_ref_model=None,
        detection_cat=None,
        flux_unit="nJy",
        cat_type="prompt",
    ):
        if not isinstance(model, ImageModel | MosaicModel):
            raise ValueError("The input model must be an ImageModel or MosaicModel.")

        self.model = model  # input model is background-subtracted
        self.segment_img = segment_img
        self.convolved_data = convolved_data
        self.kernel_fwhm = kernel_fwhm
        self.fit_psf = fit_psf
        self.mask = mask
        self.psf_ref_model = psf_ref_model
        self.detection_cat = detection_cat
        self.flux_unit = flux_unit
        self.cat_type = cat_type

        self.n_sources = len(segment_img.labels)
        self.wcs = self.model.meta.wcs
        self.meta = {}
        self.aperture_cat = None

        # define flux unit conversion factors
        if "photometry" in self.model.meta:
            self.l2_to_sb = self.model.meta.photometry.conversion_megajanskys
        else:
            self.l2_to_sb = 1.0
        self.sb_to_flux = (1.0 * (u.MJy / u.sr) * self._pixel_area).to(
            u.Unit(self.flux_unit)
        )

        if self.fit_psf and self.psf_ref_model is None:
            log.error(
                "PSF fitting is requested but no PSF reference model is provided. Skipping PSF photometry."
            )
            self.fit_psf = False

    def __len__(self):
        return self.n_sources

    def convert_l2_to_sb(self):
        """
        Convert level-2 data from units of DN/s to MJy/sr (surface
        brightness).
        """
        # the conversion in done in-place to avoid making copies of the data;
        # use dictionary syntax to set the value to avoid on-the-fly validation
        self.model["data"] *= self.l2_to_sb
        self.model["err"] *= self.l2_to_sb
        if self.convolved_data is not None:
            self.convolved_data *= self.l2_to_sb

    def convert_sb_to_flux_density(self):
        """
        Convert level-3 data from units of MJy/sr (surface brightness)
        to flux density units.

        The flux density unit is defined by self.flux_unit.
        """
        # the conversion in done in-place to avoid making copies of the data;
        # use dictionary syntax to set the value to avoid on-the-fly validation
        self.model["data"] *= self.sb_to_flux.value
        self.model["data"] <<= self.sb_to_flux.unit
        self.model["err"] *= self.sb_to_flux.value
        self.model["err"] <<= self.sb_to_flux.unit
        if self.convolved_data is not None:
            self.convolved_data *= self.sb_to_flux.value
            self.convolved_data <<= self.sb_to_flux.unit

    @lazyproperty
    def _pixscale_angle(self):
        """
        The pixel scale in arcseconds and the angle in degrees measured
        counterclockwise from the positive x axis to the "North" axis of
        the celestial coordinate system.

        The pixel is returns as a Quantity in arcsec and the angle is
        returned as a Quantity in degrees.

        Both are measured at the center of the image.
        """
        ysize, xsize = self.model.data.shape
        ycen = (ysize - 1) / 2.0
        xcen = (xsize - 1) / 2.0
        skycoord = self.wcs.pixel_to_world(xcen, ycen)
        _, pixscale, angle = pixel_scale_angle_at_skycoord(skycoord, self.wcs)
        return pixscale, angle

    @lazyproperty
    def _pixel_scale(self):
        """
        The pixel scale (as a Quantity in arcseconds) at the center of
        the image.
        """
        return self._pixscale_angle[0]

    @lazyproperty
    def _wcs_angle(self):
        """
        The angle (as a Quantity in degrees) measured counterclockwise
        from the positive x axis to the "North" axis of the celestial
        coordinate system.

        Measured at the center of the image.
        """
        return self._pixscale_angle[1]

    @lazyproperty
    def _pixel_area(self):
        """
        The pixel area (as a Quantity in steradians).

        If the meta.photometry.pixel_area value is a negative
        placeholder (e.g., -999999 sr), the value is calculated from the
        WCS at the center of the image.
        """
        if (
            "photometry" in self.model.meta
            and self.model.meta.photometry.pixel_area > 0
        ):
            pixel_area = self.model.meta.photometry.pixel_area * u.sr
        else:
            pixel_area = (self._pixel_scale**2).to(u.sr)
        return pixel_area

    @lazyproperty
    def _xypos(self):
        """
        The (x, y) source positions.

        If a detection catalog is input, then the detection catalog
        centroids are used. Otherwise, the segment catalog centroids are
        used.
        """
        if self.detection_cat is None:
            xycen = (self.x_centroid, self.y_centroid)
        else:
            xycen = (self.detection_cat.x_centroid, self.detection_cat.y_centroid)
        return np.transpose(xycen)

    @lazyproperty
    def _xypos_finite(self):
        """
        The (x, y) source positions where non-finite values have been
        replaced by to a large negative value.

        This is used with functions that fail with non-finite centroid
        values.

        For aperture photometry, at this position the aperture will not
        overlap the data, thus returning NaN fluxes and errors.
        """
        xypos = self._xypos.copy()
        nanmask = ~np.isfinite(xypos)
        xypos[nanmask] = -1000.0
        return xypos

    def calc_segment_properties(self):
        """
        Calculate the segment-based properties provided by
        `~photutils.segmentation.SourceCatalog`.

        The results are set as dynamic attributes on the class instance.
        """
        segment_cat = SegmentCatalog(
            self.model,
            self.segment_img,
            self.convolved_data,
            self._pixel_area,
            self._wcs_angle,
            detection_cat=self.detection_cat,
            flux_unit=self.flux_unit,
        )

        self.meta.update(segment_cat.meta)
        for name in segment_cat.names:
            setattr(self, name, getattr(segment_cat, name))

        # needed for detection_cat
        self.segment_cat = segment_cat

    def calc_aperture_photometry(self):
        """
        Calculate aperture photometry.

        The results are set as dynamic attributes on the class instance.
        """
        aperture_cat = ApertureCatalog(
            self.model,
            self._pixel_scale,
            self._xypos_finite,
        )
        for name in aperture_cat.names:
            setattr(self, name, getattr(aperture_cat, name))

        # needed to get aperture flux column names and descriptions
        self.aperture_cat = aperture_cat

    def calc_psf_photometry(self):
        """
        Perform PSF photometry on the sources.

        The results are set as dynamic attributes on the class instance.
        """
        psf_cat = PSFCatalog(self.model, self.psf_ref_model, self._xypos, self.mask)
        for name in psf_cat.names:
            setattr(self, name, getattr(psf_cat, name))

    def calc_daofind_properties(self):
        """
        Calculate the DAOFind sharpness and roundness1 statistics.

        The results are set as dynamic attributes on the class instance.
        """
        daofind_cat = DAOFindCatalog(
            self.model.data, self._xypos_finite, self.kernel_fwhm
        )
        for name in daofind_cat.names:
            setattr(self, name, getattr(daofind_cat, name))

    def calc_nn_properties(self):
        """
        Calculate the nearest neighbor properties.

        The results are set as dynamic attributes on the class instance.
        """
        nn_cat = NNCatalog(
            self.label, self._xypos, self._xypos_finite, self._pixel_scale
        )
        for name in nn_cat.names:
            setattr(self, name, getattr(nn_cat, name))

    @lazyproperty
    def flagged_spatial_index(self):
        """
        The spatial index bit flag encoding the projection, skycell,
        and (x, y) pixel coordinate of the source and whether the
        object is inside the skycell core region.
        """
        return np.zeros(self.n_sources, dtype=np.int64)

    @lazyproperty
    def ra(self):
        """
        The best estimate of the right ascension (ICRS).
        """
        # FIXME: just return ra_centroid as a placeholder
        return self.ra_centroid.copy()

    @lazyproperty
    def dec(self):
        """
        The best estimate of the declination (ICRS).
        """
        # FIXME: just return dec_centroid as a placeholder
        return self.dec_centroid.copy()

    @lazyproperty
    def is_extended(self):
        """
        Boolean indicating whether the source is extended.
        """
        # TODO: replace with filter-dependent ee_ratio from reference file
        # ee_ratio = ee_fraction_04 / ee_fraction_02
        ee_ratio = 1.4  # F213 placeholder value
        return self.aper04_flux > (self.aper02_flux * 1.1 * ee_ratio)

    @lazyproperty
    def warning_flags(self):
        """
        Data quality bit flags (0 = good).

        Currently sets the bits:

        * 1:
          - L2: sources whose rounded centroid pixel is not finite or has
                DO_NOT_USE set in the model DQ
          - L3: sources whose rounded centroid pixel is not finite or has
                a weight of 0
        """
        xymask = np.isfinite(self._xypos[:, 0]) & np.isfinite(self._xypos[:, 1])
        xyidx = np.round(self._xypos[xymask, :]).astype(int)
        flags = np.full(self._xypos.shape[0], 0, dtype=np.int32)

        # sources whose centroid pixel is not finite
        flags[~xymask] = pixel.DO_NOT_USE

        try:
            # L2 images have a dq array
            dqflags = self.model.dq[xyidx[:, 1], xyidx[:, 0]]
            # if dqflags contains the DO_NOT_USE flag, set to DO_NOT_USE
            # (dq=1), otherwise 0
            flags[xymask] = dqflags & pixel.DO_NOT_USE

        except AttributeError:
            # L3 images
            mask = self.model.weight == 0
            flags[xymask] = mask[xyidx[:, 1], xyidx[:, 0]].astype(np.int32)

        return flags

    @lazyproperty
    def image_flags(self):
        """
        Data quality bit flag.

        Non-zero if a pixel within the segment was flagged in one of the
        input images.
        """
        return np.zeros(self.n_sources, dtype=np.int32)

    def update_metadata(self):
        """
        Update the metadata dictionary with the package version
        information and aperture parameters.
        """
        ver_key = "versions"
        if "version" in self.meta:
            # depends on photutils version
            self.meta[ver_key] = self.meta.pop("version")

        ver_dict = self.meta.get("versions", None)
        if ver_dict is not None:
            ver_dict["romancal"] = romancal_version
            packages = [
                "Python",
                "numpy",
                "scipy",
                "astropy",
                "photutils",
                "gwcs",
                "romancal",
            ]
            ver_dict = {key: ver_dict[key] for key in packages if key in ver_dict}
            self.meta[ver_key] = ver_dict

        # reorder the metadata dictionary
        self.meta.pop("localbkg_width")
        keys = ["date", "versions", "apermask_method", "kron_params"]
        old_meta = self.meta.copy()
        self.meta = {key: old_meta[key] for key in keys}

        # reformat the aperture radii for the metadata to remove
        # Quantity objects
        if self.aperture_cat is not None:
            aper_radii = self.aperture_cat.aperture_radii.copy()
            aper_radii["circle_arcsec"] = aper_radii.pop("circle").value
            aper_radii["annulus_arcsec"] = aper_radii.pop("annulus").value
            self.meta["aperture_radii"] = aper_radii

    @lazyproperty
    def column_descriptions(self):
        """
        A dictionary of the output catalog column descriptions.

        The order is not important.
        """
        col = {}
        col["label"] = "Label of the source segment in the segmentation image"
        col["flagged_spatial_index"] = (
            "Bit flag encoding the overlap flag, projection, skycell, and "
            "pixel coordinates of the source"
        )

        col["x_centroid"] = (
            "Column coordinate of the source centroid in the detection "
            "image from image moments (0 indexed)"
        )
        col["y_centroid"] = (
            "Row coordinate of the source centroid in the detection image "
            "from image moments (0 indexed)"
        )
        col["x_centroid_win"] = (
            "Column coordinate of the windowed source centroid in the "
            "detection image from image moments (0 indexed)"
        )
        col["y_centroid_win"] = (
            "Row coordinate of the windowed source centroid in the "
            "detection image from image moments (0 indexed)"
        )

        col["ra"] = "Best estimate of the right ascension (ICRS)"
        col["dec"] = "Best estimate of the declination (ICRS)"
        col["ra_centroid"] = "Right ascension (ICRS) of the source centroid"
        col["dec_centroid"] = "Declination (ICRS) of the source centroid"
        col["ra_centroid_win"] = (
            "Right ascension (ICRS) of the windowed source centroid"
        )
        col["dec_centroid_win"] = "Declination (ICRS) of the windowed source centroid"
        col["ra_psf"] = "Right ascension (ICRS) of the PSF-fitted position"
        col["dec_psf"] = "Declination (ICRS) of the PSF-fitted position"

        col["bbox_xmin"] = (
            "Column index of the left edge of the source bounding box (0 indexed)"
        )
        col["bbox_xmax"] = (
            "Column index of the right edge of the source bounding box (0 indexed)"
        )
        col["bbox_ymin"] = (
            "Row index of the bottom edge of the source bounding box (0 indexed)"
        )
        col["bbox_ymax"] = (
            "Row index of the top edge of the source bounding box (0 indexed)"
        )
        col["semimajor"] = (
            "Length of the source semimajor axis computed from image moments"
        )
        col["semiminor"] = (
            "Length of the source semiminor axis computed from image moments"
        )
        col["fwhm"] = (
            "Circularized full width at half maximum (FWHM) "
            "calculated from the semimajor and semiminor axes "
            "as 2*sqrt(ln(2) * (semimajor**2 + semiminor**2))"
        )
        col["ellipticity"] = "Source ellipticity as 1 - (semimajor / semiminor)"
        col["orientation_pix"] = (
            "Angle measured counter-clockwise from the positive X axis to the source major axis computed from image moments"
        )
        col["orientation_sky"] = (
            "Position angle from North of the source major axis computed from image moments"
        )

        col["cxx"] = (
            "Coefficient for the x**2 term in the generalized quadratic ellipse equation"
        )
        col["cxy"] = (
            "Coefficient for the x*y term in the generalized quadratic ellipse equation"
        )
        col["cyy"] = (
            "Coefficient for the y**2 term in the generalized quadratic ellipse equation"
        )

        col["segment_flux"] = "Isophotal flux"
        col["segment_area"] = "Area of the source segment"
        col["kron_radius"] = "Unscaled first-moment Kron radius"
        col["fluxfrac_radius_50"] = (
            "Radius of a circle centered on the source centroid that encloses 50% of the Kron flux"
        )
        col["kron_flux"] = "Flux within the elliptical Kron aperture"
        col["kron_abmag"] = "AB magnitude within the elliptical Kron aperture"
        col["aper_bkg_flux"] = "Local background estimated within a circular annulus"

        col["x_psf"] = "Column coordinate of the source from PSF fitting (0 indexed)"
        col["y_psf"] = "Row coordinate of the source from PSF fitting (0 indexed)"
        col["psf_flux"] = "Total PSF flux"
        col["psf_gof"] = "PSF goodness of fit metric"
        col["psf_flags"] = "PSF fitting bit flags (0 = good)"

        col["warning_flags"] = "Warning bit flags (0 = good)"
        col["image_flags"] = "Image quality bit flags (0 = good)"

        col["is_extended"] = (
            "Flag indicating that the source appears to be more extended than a point source"
        )
        col["sharpness"] = "Photutils DAOStarFinder sharpness statistic"
        col["roundness1"] = "Photutils DAOStarFinder roundness1 statistic"
        col["nn_label"] = "Segment label of the nearest neighbor in this skycell"
        col["nn_distance"] = "Distance to the nearest neighbor in this skycell"

        # add the aperture flux column descriptions
        if self.aperture_cat is not None:
            col.update(self.aperture_cat.aperture_flux_descriptions)

        # add the "*_err" column descriptions
        for column in self.column_names:
            if column.endswith("_err"):
                base_column = column.replace("_err", "")
                col[column] = f"Uncertainty in {base_column}"

        return col

    @lazyproperty
    def aper_colnames(self):
        """
        An ordered list of the aperture column names.
        """
        # define the aperture background flux column names
        aper_colnames = [
            "aper_bkg_flux",
            "aper_bkg_flux_err",
        ]

        # define the aperture flux column names,
        # e.g, aper01_flux, aper01_flux_err, etc.
        for colname in self.aperture_cat.aperture_flux_colnames:
            aper_colnames.append(colname)
            aper_colnames.append(f"{colname}_err")

        return aper_colnames

    @lazyproperty
    def flux_colnames(self):
        """
        An ordered list of the flux column names.
        """
        other_colnames = [
            "segment_flux",
            "segment_flux_err",
            "kron_flux",
            "kron_flux_err",
            "kron_abmag",
            "kron_abmag_err",
        ]
        psf_colnames = ["psf_flux", "psf_flux_err"]

        if self.cat_type in ("prompt", "forced_full", "dr_band"):
            flux_colnames = self.aper_colnames
            if self.fit_psf:
                flux_colnames.extend(psf_colnames)
            flux_colnames.extend(other_colnames)

        elif self.cat_type in ("dr_det", "forced_det"):
            flux_colnames = []

        else:
            raise ValueError(f"Unknown catalog type: {self.cat_type}")

        return flux_colnames

    @lazyproperty
    def column_names(self):
        """
        An ordered list of the output catalog column names.

        This list determines which values are calculated in the output
        catalog.
        """
        base_colnames = [
            "label",
            "flagged_spatial_index",
            "x_centroid",
            "y_centroid",
            "x_centroid_err",
            "y_centroid_err",
        ]
        xywin_colnames = [
            "x_centroid_win",
            "y_centroid_win",
            "x_centroid_win_err",
            "y_centroid_win_err",
        ]
        skybest_colnames = [
            "ra",
            "dec",
        ]
        sky_colnames = [
            "ra_centroid",
            "dec_centroid",
            "ra_centroid_err",
            "dec_centroid_err",
        ]
        skywin_colnames = [
            "ra_centroid_win",
            "dec_centroid_win",
            "ra_centroid_win_err",
            "dec_centroid_win_err",
        ]
        skypsf_colnames = [
            "ra_psf",
            "dec_psf",
            "ra_psf_err",
            "dec_psf_err",
        ]
        segm_colnames = [
            "bbox_xmin",
            "bbox_xmax",
            "bbox_ymin",
            "bbox_ymax",
            "segment_area",
        ]
        shape_colnames = [
            "semimajor",
            "semiminor",
            "fwhm",
            "ellipticity",
            "orientation_pix",
            "orientation_sky",
            "cxx",
            "cxy",
            "cyy",
            "kron_radius",
        ]
        nn_colnames = [
            "nn_label",
            "nn_distance",
        ]
        othershape_colnames = [
            "sharpness",
            "roundness1",
            "is_extended",
            "fluxfrac_radius_50",
        ]
        xypsf_colnames = [
            "x_psf",
            "x_psf_err",
            "y_psf",
            "y_psf_err",
        ]
        flag_columns = [
            "warning_flags",
            "image_flags",
        ]
        psf_flags_colnames = [
            "psf_flags",
            "psf_gof",
        ]

        det_colnames = []
        det_colnames.extend(segm_colnames)
        det_colnames.extend(shape_colnames)
        det_colnames.extend(nn_colnames)

        # These are band-specific columns for the multiband catalog.
        # They are also used for the forced catalog.
        band_colnames = ["label"]  # needed to join the filter catalogs
        if self.fit_psf:
            band_colnames.extend(xypsf_colnames)
            band_colnames.extend(skypsf_colnames)
            band_colnames.extend(psf_flags_colnames)
        band_colnames.extend(othershape_colnames)
        band_colnames.extend(self.flux_colnames)
        self.band_colnames = band_colnames

        if self.cat_type in ("prompt", "forced_full"):
            colnames = []
            colnames.extend(base_colnames)
            colnames.extend(xywin_colnames)
            if self.fit_psf:
                colnames.extend(xypsf_colnames)
            colnames.extend(skybest_colnames)
            colnames.extend(sky_colnames)
            colnames.extend(skywin_colnames)
            if self.fit_psf:
                colnames.extend(skypsf_colnames)
            colnames.extend(det_colnames)
            colnames.extend(othershape_colnames)
            colnames.extend(self.flux_colnames)

            colnames.extend(flag_columns)
            if self.fit_psf:
                colnames.extend(psf_flags_colnames)

        elif self.cat_type == "forced_det":
            colnames = ["label"]  # needed to join the forced catalogs
            colnames.extend(base_colnames)
            colnames.extend(skybest_colnames)
            colnames.extend(sky_colnames)
            colnames.extend(shape_colnames)
            colnames.extend(nn_colnames)

        elif self.cat_type == "dr_det":
            colnames = []
            colnames.extend(base_colnames)
            colnames.extend(xywin_colnames)
            colnames.extend(skybest_colnames)
            colnames.extend(sky_colnames)
            colnames.extend(skywin_colnames)
            colnames.extend(det_colnames)
            colnames.extend(flag_columns)

        elif self.cat_type == "dr_band":
            colnames = band_colnames.copy()

        return colnames

    def _prefix_forced(self, catalog):
        """
        Prefix select columns of the catalog with "forced_" for the
        forced catalog.

        Parameters
        ----------
        catalog : `~astropy.table.Table`
            The source catalog to prefix.

        Returns
        -------
        catalog : `~astropy.table.Table`
            The updated source catalog.
        """
        # prefix all columns (except "label") with "forced_"
        if self.cat_type == "forced_det":
            for colname in catalog.colnames:
                if colname != "label":
                    catalog.rename_column(colname, f"forced_{colname}")

        # prefix the self.band_colnames with "forced_"
        if self.cat_type == "forced_full":
            for colname in [*self.band_colnames, "warning_flags"]:
                if colname in catalog.colnames and colname != "label":
                    catalog.rename_column(colname, f"forced_{colname}")

        return catalog

    @lazyproperty
    def catalog(self):
        """
        The final source catalog as an Astropy Table.
        """
        # convert data to flux units
        if isinstance(self.model, ImageModel):
            self.convert_l2_to_sb()
        self.convert_sb_to_flux_density()

        # make measurements - the order of these calculations is important
        log.info("Calculating segment properties")
        self.calc_segment_properties()

        # NOTE: we cannot access self.column_names before
        # calc_aperture_photometry is called because the aperture columns
        # names are dynmically generated
        if self.cat_type != "forced_det":
            log.info("Calculating aperture photometry")
            self.calc_aperture_photometry()

        daofind_cols = {"sharpness", "roundness1"}
        if daofind_cols & set(self.column_names):
            log.info("Calculating DAOFind properties")
            self.calc_daofind_properties()

        if any("nn_" in col for col in self.column_names):
            log.info("Calculating nearest neighbor properties")
            self.calc_nn_properties()

        if any("psf" in col for col in self.column_names):
            # TODO: compute force_full PSF photometry at forced positions
            log.info("Calculating PSF photometry")
            self.calc_psf_photometry()

        # put the measurements into a Table
        catalog = QTable()
        for column in self.column_names:
            catalog[column] = getattr(self, column)
            descrip = self.column_descriptions.get(column, None)
            catalog[column].info.description = descrip
        self.update_metadata()
        catalog.meta.update(self.meta)

        # prefix select columns with "forced_" for the forced catalog
        catalog = self._prefix_forced(catalog)

        # convert QTable to Table to avoid having Quantity columns
        catalog = Table(catalog)

        return catalog



"""
Module for the multiband source catalog step.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from astropy.table import join
from roman_datamodels import datamodels

from romancal.datamodels import ModelLibrary
from romancal.multiband_catalog.background import subtract_background_library
from romancal.multiband_catalog.detection_image import make_detection_image
from romancal.multiband_catalog.utils import add_filter_to_colnames
from romancal.source_catalog.background import RomanBackground
from romancal.source_catalog.detection import make_segmentation_image
from romancal.source_catalog.save_utils import save_all_results, save_empty_results
from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.source_catalog.utils import copy_mosaic_meta
from romancal.stpipe import RomanStep

if TYPE_CHECKING:
    from typing import ClassVar

__all__ = ["MultibandCatalogStep"]

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class MultibandCatalogStep(RomanStep):
    """
    Create a multiband catalog of sources including photometry and basic
    shape measurements.

    Parameters
    -----------
    input : str or `~romancal.datamodels.ModelLibrary`
        Path to an ASDF file or a `~romancal.datamodels.ModelLibrary`
        that contains `~roman_datamodels.datamodels.MosaicImageModel`
        models.
    """

    class_alias = "multiband_catalog"
    reference_file_types: ClassVar = []

    spec = """
        bkg_boxsize = integer(default=100)   # background mesh box size in pixels
        kernel_fwhms = float_list(default=None)  # Gaussian kernel FWHM in pixels
        snr_threshold = float(default=3.0)    # per-pixel SNR threshold above the bkg
        npixels = integer(default=25)         # min number of pixels in source
        deblend = boolean(default=False)      # deblend sources?
        suffix = string(default='cat')        # Default suffix for output files
        fit_psf = boolean(default=True)       # fit source PSFs for accurate astrometry?
    """

    def process(self, library):
        # All input MosaicImages in the ModelLibrary are assumed to have
        # the same shape and be pixel aligned.
        if isinstance(library, str):
            library = ModelLibrary(library)
        if not isinstance(library, ModelLibrary):
            raise TypeError("library input must be a ModelLibrary object")

        with library:
            example_model = library.borrow(0)
            library.shelve(example_model, modify=False)

        # Initialize the source catalog model, copying the metadata
        # from the input model
        cat_model = datamodels.MultibandSourceCatalogModel.create_minimal(
            {"meta": example_model.meta}
        )
        if isinstance(example_model, datamodels.MosaicModel):
            copy_mosaic_meta(example_model, cat_model)
        else:
            cat_model.meta.optical_element = (
                example_model.meta.instrument.optical_element
            )

        # Define the output filename for the source catalog model
        try:
            cat_model.meta.filename = library.asn["products"][0]["name"]
        except (AttributeError, KeyError):
            cat_model.meta.filename = "multiband_catalog"

        log.info("Calculating and subtracting background")
        library = subtract_background_library(library, self.bkg_boxsize)

        log.info("Creating detection image")
        # Define the kernel FWHMs for the detection image
        # TODO: sensible defaults
        # TODO: redefine in terms of intrinsic FWHM
        if self.kernel_fwhms is None:
            self.kernel_fwhms = [2.0, 20.0]

        # TODO: det_img is saved in the MosaicSegmentationMapModel;
        # do we also want to save the det_err?
        det_img, det_err = make_detection_image(library, self.kernel_fwhms)

        # Estimate background rms from detection image to calculate a
        # threshold for source detection
        mask = ~np.isfinite(det_img) | ~np.isfinite(det_err) | (det_err <= 0)

        # Return an empty segmentation image and catalog table if all
        # pixels are masked in the detection image.
        if np.all(mask):
            msg = "Cannot create source catalog. All pixels in the detection image are masked."
            return save_empty_results(self, det_img.shape, cat_model, msg=msg)

        bkg = RomanBackground(
            det_img,
            box_size=self.bkg_boxsize,
            coverage_mask=mask,
        )
        bkg_rms = bkg.background_rms

        log.info("Detecting sources")
        segment_img = make_segmentation_image(
            det_img,
            snr_threshold=self.snr_threshold,
            npixels=self.npixels,
            bkg_rms=bkg_rms,
            deblend=self.deblend,
            mask=mask,
        )

        if segment_img is None:  # no sources found
            msg = "Cannot create source catalog. No sources were detected."
            return save_empty_results(self, det_img.shape, cat_model, msg=msg)

        segment_img.detection_image = det_img

        # Define the detection image model
        det_model = datamodels.MosaicModel()
        det_model.data = det_img
        det_model.err = det_err

        # TODO: this is a temporary solution to get model attributes
        # currently needed in RomanSourceCatalog
        det_model.weight = example_model.weight
        det_model.meta = example_model.meta

        # The stellar FWHM is needed to define the kernel used for
        # the DAOStarFinder sharpness and roundness properties.
        # TODO: measure on a secondary detection image with minimal
        # smoothing?; use the same detection image for basic shape
        # measurements?
        star_kernel_fwhm = np.min(self.kernel_fwhms)

        log.info("Creating catalog for detection image")
        det_catobj = RomanSourceCatalog(
            det_model,
            segment_img,
            det_img,
            star_kernel_fwhm,
            fit_psf=self.fit_psf,
            detection_cat=None,
            mask=mask,
            cat_type="dr_det",
        )

        # Generate the catalog for the detection image.
        # We need to make this catalog before we pass det_catobj
        # to the RomanSourceCatalog constructor.
        det_cat = det_catobj.catalog

        # Create catalogs for each input image
        with library:
            for model in library:
                mask = (
                    ~np.isfinite(model.data)
                    | ~np.isfinite(model.err)
                    | (model.err <= 0)
                )

                if self.fit_psf:
                    filter_name = model.meta.instrument.optical_element
                    log.info(f"Creating catalog for {filter_name} image")
                    ref_file = self.get_reference_file(model, "epsf")
                    log.info("Using ePSF reference file: %s", ref_file)
                    psf_ref_model = datamodels.open(ref_file)
                else:
                    psf_ref_model = None

                catobj = RomanSourceCatalog(
                    model,
                    segment_img,
                    None,
                    star_kernel_fwhm,
                    fit_psf=self.fit_psf,
                    detection_cat=det_catobj,
                    mask=mask,
                    psf_ref_model=psf_ref_model,
                    cat_type="dr_band",
                )

                # Add the filter name to the column names
                filter_name = model.meta.instrument.optical_element
                cat = add_filter_to_colnames(catobj.catalog, filter_name)

                # TODO: what metadata do we want to keep, if any,
                # from the filter catalogs?
                cat.meta = None

                # Add the filter catalog to the multiband catalog.
                # The outer join prevents an empty table if any
                # columns have the same name but different values
                # (e.g., repeated filter names)
                det_cat = join(det_cat, cat, keys="label", join_type="outer")
                library.shelve(model, modify=False)

        # Put the resulting multiband catalog in the model
        cat_model.source_catalog = det_cat

        return save_all_results(self, segment_img, cat_model)
